"""Test module for ecl2df.pillars"""

from pathlib import Path

import pandas as pd
import pytest

from ecl2df import ecl2csv, grid, pillars
from ecl2df.eclfiles import EclFiles

TESTDIR = Path(__file__).absolute().parent
REEK = str(TESTDIR / "data/reek/eclipse/model/2_R001_REEK-0.DATA")


def test_pillars():
    """Test that we can build a dataframe of pillar statistics"""
    eclfiles = EclFiles(REEK)
    pillars_df = pillars.df(eclfiles)
    assert "PILLAR" in pillars_df
    assert "VOLUME" in pillars_df
    assert "PORV" in pillars_df
    assert "PERMX" in pillars_df
    assert "X" in pillars_df
    assert "Y" in pillars_df
    assert "PORO" in pillars_df
    assert "OILVOL" not in pillars_df
    assert "FIPNUM" not in pillars_df
    assert "EQLNUM" not in pillars_df
    assert "OWC" not in pillars_df
    assert "GOC" not in pillars_df
    assert len(pillars_df) == 2560

    pillars_df = pillars.df(eclfiles, region="FIPNUM")
    assert "FIPNUM" in pillars_df
    assert len(pillars_df["FIPNUM"].unique()) == 6
    assert "OILVOL" not in pillars_df

    pillars_df = pillars.df(eclfiles, rstdates="first")
    firstdate = str(grid.dates2rstindices(eclfiles, "first")[1][0])
    assert "OILVOL@" + firstdate in pillars_df
    assert "GASVOL@" + firstdate in pillars_df
    assert "WATVOL@" + firstdate in pillars_df

    pillars_df = pillars.df(eclfiles, rstdates="last", soilcutoff=0.2, sgascutoff=0.2)
    lastdate = str(grid.dates2rstindices(eclfiles, "last")[1][0])
    assert "OWC@" + lastdate in pillars_df
    assert "GOC@" + lastdate not in pillars_df  # Because the dataset has no GAS...

    # Grouping by unknowns only trigger a warning
    pd.testing.assert_frame_equal(
        pillars.df(eclfiles), pillars.df(eclfiles, region="FOOBAR")
    )


# These frame will be reused in the following test.
PILLARS_WITH_UPFLANK = pd.DataFrame(
    columns=["PILLAR", "SWAT", "SOIL", "Z"],
    data=[
        ["1-1", 0.2, 0.8, 950],
        ["1-1", 0.7, 0.3, 951],
        ["1-1", 0.9, 0.1, 952],
        ["1-1", 1, 0, 953],
        ["2-1", 0.2, 1, 400],  # Upflank oil, to be ignored.
    ],
)
PILLARS_GAS_IN_WATER = pd.DataFrame(
    columns=["PILLAR", "SWAT", "SOIL", "SGAS", "Z"],
    data=[
        ["1-1", 0.2, 0.02, 0.8, 940],
        ["1-1", 0.2, 0.1, 0.7, 942],
        ["1-1", 0.2, 0.4, 0.4, 945],
        ["1-1", 0.2, 0.8, 0, 950],
        ["1-1", 0.7, 0.3, 0, 951],
        ["1-1", 0.9, 0.1, 0, 952],
        ["1-1", 1, 0, 0, 953],
        # Add a row with gas saturation in water, this
        # could be due to a gas injector and
        # should not be picked up as a GOC:
        ["1-1", 0.5, 0, 0.5, 953],
        ["1-1", 1, 0, 0, 953],
    ],
)


@pytest.mark.parametrize(
    "pillar_df, args, expectedrows",
    [
        (pd.DataFrame(), {}, []),
        pytest.param(
            pd.DataFrame([{"SWAT": 1}]),
            {},
            None,
            marks=pytest.mark.xfail(raises=KeyError),
            id="no_pillar_column",
        ),
        pytest.param(
            pd.DataFrame([{"SWAT": 1, "I": 1, "J": 1}]),
            {},
            None,
            marks=pytest.mark.xfail(raises=KeyError),
            id="no_z_column",
        ),
        pytest.param(
            pd.DataFrame([{"SWAT": 1, "I": 1, "J": 1, "Z": 1000}]),
            {},
            [],
            id="no_saturations_gives_empty_result",
        ),
        (
            pd.DataFrame(
                columns=["PILLAR", "SWAT", "SOIL", "SGAS", "Z"],
                data=[["1-1", 1, 0, 0, 1000]],
            ),
            {},
            [],
        ),
        (
            pd.DataFrame(
                columns=["PILLAR", "SWAT", "SOIL", "SGAS", "Z"],
                data=[["1-1", 1, 0, 0, 1000], ["1-1", 0.5, 0.5, 0, 999]],
            ),
            {},
            [{"PILLAR": "1-1", "OWC": 999}],
        ),
        (
            pd.DataFrame(
                columns=["PILLAR", "SWAT", "SOIL", "SGAS", "Z"],
                data=[["1-1", 1, 0, 0, 1000], ["1-1", 0.5, 0.5, 0, 999]],
            ),
            {"soilcutoff": 0.2},
            [{"PILLAR": "1-1", "OWC": 999}],
        ),
        pytest.param(
            pd.DataFrame(
                columns=["PILLAR", "SWAT", "SOIL", "SGAS", "Z"],
                data=[["1-1", 1, 0, 0, 1000], ["1-1", 0.5, 0.5, 0, 999]],
            ),
            {"soilcutoff": 0.6},
            [],
            id="bump_soilcutoff_giving_no_contact",
        ),
        pytest.param(
            pd.DataFrame(
                columns=["PILLAR", "SWAT", "SGAS", "Z"],
                data=[["1-1", 1, 0, 1000], ["1-1", 0.2, 0.8, 999]],
            ),
            {},
            [{"PILLAR": "1-1", "GWC": 999}],
            id="two-phase_gas-water",
        ),
        pytest.param(
            pd.DataFrame(
                columns=["PILLAR", "SWAT", "SOIL", "Z"],
                data=[["1-1", 1, 0, 1000], ["1-1", 0.2, 0.8, 999]],
            ),
            {},
            [{"PILLAR": "1-1", "OWC": 999}],
            id="two-phase-oil-water",
        ),
        pytest.param(
            pd.DataFrame(
                columns=["PILLAR", "EQLNUM", "SWAT", "SOIL", "Z"],
                data=[
                    ["1-1", 1, 1, 0, 1000],
                    ["1-1", 1, 0.2, 0.8, 999],
                    ["1-1", 2, 1, 0, 2000],
                    ["1-1", 2, 0.2, 0.8, 1999],
                ],
            ),
            {"region": "EQLNUM"},
            [
                {"PILLAR": "1-1", "EQLNUM": 1, "OWC": 999},
                {"PILLAR": "1-1", "EQLNUM": 2, "OWC": 1999},
            ],
            id="testing_region_support",
        ),
        pytest.param(
            PILLARS_WITH_UPFLANK,
            {},
            [
                {"PILLAR": "1-1", "OWC": 951},
            ],
            id="upflank_oil_ignored",
        ),
        pytest.param(
            PILLARS_WITH_UPFLANK,
            {"swatcutoff": 0.05},
            [
                {"PILLAR": "1-1", "OWC": 951},
                {"PILLAR": "2-1", "OWC": 400},
            ],
            id="swatcutoff_includes_upflank",
        ),
        # "OWC is deepest cell centre with saturation more than soilcutoff":
        pytest.param(
            PILLARS_WITH_UPFLANK,
            {"soilcutoff": 0.05},
            [
                {"PILLAR": "1-1", "OWC": 952},
            ],
            id="soilcutoff_0.05",
        ),
        pytest.param(
            PILLARS_WITH_UPFLANK,
            {"soilcutoff": 0.25},
            [
                {"PILLAR": "1-1", "OWC": 951},
            ],
            id="soilcutoff_0.25",
        ),
        pytest.param(
            PILLARS_WITH_UPFLANK,
            {"soilcutoff": 0.7},
            [
                {"PILLAR": "1-1", "OWC": 950},
            ],
            id="soilcutoff_0.7",
        ),
        pytest.param(
            PILLARS_WITH_UPFLANK,
            {"soilcutoff": 0.8},
            [],
            id="soilcutoff_0.8",
        ),
        # "GOC is deepest point with gas saturation is more than sgascutoff,
        # and where some cells have oil saturation more than gocsoilcutoff"
        pytest.param(
            PILLARS_GAS_IN_WATER,
            {},
            [{"PILLAR": "1-1", "OWC": 951, "GOC": 940}],
            id="goc_gas_in_water",
        ),
        pytest.param(
            PILLARS_GAS_IN_WATER,
            {"sgascutoff": 0.05},
            [{"PILLAR": "1-1", "OWC": 951, "GOC": 945}],
            id="sgascutoff_0.05",
        ),
        pytest.param(
            PILLARS_GAS_IN_WATER,
            {"sgascutoff": 0.4},
            [{"PILLAR": "1-1", "OWC": 951, "GOC": 942}],
            id="sgascutoff_0.4",
        ),
        pytest.param(
            PILLARS_GAS_IN_WATER,
            {"sgascutoff": 0.75},
            [{"PILLAR": "1-1", "OWC": 951, "GOC": 940}],
            id="sgascutoff_0.75",
        ),
    ],
)
def test_compute_pillar_contacts(pillar_df, args, expectedrows):
    pd.testing.assert_frame_equal(
        pillars.compute_pillar_contacts(pillar_df, **args), pd.DataFrame(expectedrows)
    )


@pytest.mark.parametrize(
    "dframe, datestr, expectedrows",
    [
        (pd.DataFrame(), None, []),
        (
            pd.DataFrame([{"PORV": 1, "SWAT": 0.9, "SGAS": 0}]),
            None,
            [{"SOIL": 0.1, "WATVOL": 0.9, "GASVOL": 0, "OILVOL": 0.1}],
        ),
        (
            # Empty datestring is the same as None:
            pd.DataFrame([{"PORV": 1, "SWAT": 0.9, "SGAS": 0}]),
            "",
            [{"SOIL": 0.1, "WATVOL": 0.9, "GASVOL": 0, "OILVOL": 0.1}],
        ),
        (
            # With a date included:
            pd.DataFrame([{"PORV": 1, "SWAT@2000-01-01": 0.9, "SGAS@2000-01-01": 0}]),
            "2000-01-01",
            [
                {
                    "SOIL@2000-01-01": 0.1,
                    "WATVOL@2000-01-01": 0.9,
                    "GASVOL@2000-01-01": 0,
                    "OILVOL@2000-01-01": 0.1,
                }
            ],
        ),
        (
            # Asking for a date for which there is no data:
            pd.DataFrame([{"PORV": 1, "SWAT@2000-01-01": 0.9, "SGAS@2000-01-01": 0}]),
            "2001-01-01",
            [],
        ),
        (
            # Two phase oil-water
            pd.DataFrame([{"PORV": 1, "SWAT": 0.9}]),
            None,
            [{"SOIL": 0.1, "WATVOL": 0.9, "OILVOL": 0.1}],
        ),
        (
            # Including surface conditions
            pd.DataFrame(
                [{"PORV": 1, "SWAT": 0.5, "SGAS": 0.2, "1OVERBO": 0.8, "1OVERBG": 2}]
            ),
            None,
            [
                {
                    "SOIL": 0.3,
                    "WATVOL": 0.5,
                    "GASVOL": 0.2,
                    "OILVOL": 0.3,
                    "OILVOLSURF": 0.3 * 0.8,
                    "GASVOLSURF": 0.2 * 2,
                }
            ],
        ),
    ],
)
def test_compute_volumes(dframe, datestr, expectedrows):
    pd.testing.assert_frame_equal(
        pillars.compute_volumes(dframe, datestr), pd.DataFrame(expectedrows)
    )


def test_main(tmp_path, mocker):
    """Test command line interface"""
    tmpcsvfile = tmp_path / "pillars.csv"
    mocker.patch("sys.argv", ["ecl2csv", "pillars", REEK, "-o", str(tmpcsvfile)])
    ecl2csv.main()
    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert "PILLAR" in disk_df
    assert not disk_df.empty
    assert len(disk_df) == 2560

    # Group over every pillar, no matter what FIPNUM. One single row output
    mocker.patch(
        "sys.argv",
        [
            "ecl2csv",
            "pillars",
            REEK,
            "--region",
            "",
            "--group",
            "-o",
            str(tmpcsvfile),
        ],
    )
    ecl2csv.main()
    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert "PILLAR" not in disk_df  # because of grouping
    assert "FIPNUM" not in disk_df
    assert "EQLNUM" not in disk_df
    # We are getting a single row only
    assert len(disk_df) == 1
    assert not disk_df.empty

    # Pillars pr region, but no grouping
    mocker.patch(
        "sys.argv",
        [
            "ecl2csv",
            "pillars",
            REEK,
            "--region",
            "FIPNUM",
            "-o",
            str(tmpcsvfile),
        ],
    )
    ecl2csv.main()
    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert "PILLAR" in disk_df
    assert "FIPNUM" in disk_df
    assert "EQLNUM" not in disk_df
    assert len(disk_df) == 7675

    # Group pr. FIPNUM:
    mocker.patch(
        "sys.argv",
        [
            "ecl2csv",
            "pillars",
            REEK,
            "--region",
            "FIPNUM",
            "--group",
            "-o",
            str(tmpcsvfile),
        ],
    )
    ecl2csv.main()
    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert "PILLAR" not in disk_df  # because of grouping
    assert "FIPNUM" in disk_df  # grouped by this.
    assert len(disk_df) == 6

    # Test dates:
    mocker.patch(
        "sys.argv",
        [
            "ecl2csv",
            "pillars",
            REEK,
            "--region",
            "",
            "--group",
            "--rstdates",
            "first",
            "-o",
            str(tmpcsvfile),
        ],
    )
    ecl2csv.main()
    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert "PILLAR" not in disk_df  # because of region averaging
    assert "FIPNUM" not in disk_df
    assert "WATVOL@2000-01-01" in disk_df
    assert "OILVOL@2000-01-01" in disk_df
    assert "OWC@2000-01-01" in disk_df
    # Check that we don't get meaningless temporary data:
    assert "SWAT@2000-01-01" not in disk_df
    assert "SOIL@2000-01-01" not in disk_df
    assert len(disk_df) == 1

    # Test dates:
    mocker.patch(
        "sys.argv",
        [
            "ecl2csv",
            "pillars",
            REEK,
            "--region",
            "",
            "--group",
            "--rstdates",
            "last",
            "-o",
            str(tmpcsvfile),
        ],
    )
    ecl2csv.main()
    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert "PILLAR" not in disk_df  # because of region averaging
    assert "FIPNUM" not in disk_df
    assert "WATVOL@2001-08-01" in disk_df
    assert "OILVOL@2001-08-01" in disk_df
    assert "OWC@2001-08-01" in disk_df
    # Check that we don't get meaningless temporary data:
    assert "SWAT@2001-01-01" not in disk_df
    assert "SOIL@2000-01-01" not in disk_df
    assert len(disk_df) == 1

    # Test all dates:
    mocker.patch(
        "sys.argv",
        [
            "ecl2csv",
            "pillars",
            REEK,
            "--region",
            "",
            "--group",
            "--rstdates",
            "all",
            "-o",
            str(tmpcsvfile),
        ],
    )
    ecl2csv.main()
    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert "PILLAR" not in disk_df  # because of region averaging
    assert "FIPNUM" not in disk_df
    assert "WATVOL@2001-08-01" in disk_df
    assert "WATVOL@2000-07-01" in disk_df
    assert "WATVOL@2000-01-01" in disk_df
    assert "WATVOL@2001-02-01" in disk_df
    assert len(disk_df) == 1

    # Test stacked dates:
    mocker.patch(
        "sys.argv",
        [
            "ecl2csv",
            "pillars",
            REEK,
            "--region",
            "",
            "--group",
            "--rstdates",
            "all",
            "--stackdates",
            "-o",
            str(tmpcsvfile),
        ],
    )
    ecl2csv.main()
    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert "PILLAR" not in disk_df  # because of region averaging
    assert "FIPNUM" not in disk_df
    assert "WATVOL@2001-08-01" not in disk_df
    assert "WATVOL@2000-07-01" not in disk_df
    assert "WATVOL@2000-01-01" not in disk_df
    assert "WATVOL@2001-02-01" not in disk_df
    assert "WATVOL" in disk_df
    assert "DATE" in disk_df
    assert len(disk_df) == 4

    # Test stacked dates, no grouping:
    mocker.patch(
        "sys.argv",
        [
            "ecl2csv",
            "pillars",
            REEK,
            "--region",
            "FIPNUM",
            "--rstdates",
            "all",
            "--stackdates",
            "-o",
            str(tmpcsvfile),
        ],
    )
    ecl2csv.main()
    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert "PILLAR" in disk_df
    assert "FIPNUM" in disk_df
    assert "WATVOL@2001-02-01" not in disk_df
    assert "WATVOL" in disk_df
    assert "DATE" in disk_df
    assert len(disk_df) == 30700

    # Test stacked dates but with grouping only on pillars
    mocker.patch(
        "sys.argv",
        [
            "ecl2csv",
            "pillars",
            "-v",
            REEK,
            "--rstdates",
            "all",
            "--stackdates",
            "-o",
            str(tmpcsvfile),
        ],
    )
    ecl2csv.main()
    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert "PILLAR" in disk_df
    assert "FIPNUM" not in disk_df
    assert "WATVOL@2001-08-01" not in disk_df
    assert "WATVOL@2000-07-01" not in disk_df
    assert "WATVOL@2000-01-01" not in disk_df
    assert "WATVOL@2001-02-01" not in disk_df
    assert "WATVOL" in disk_df
    assert "OILVOL" in disk_df
    assert "DATE" in disk_df
    assert len(disk_df) == 10240 == 2560 * len(disk_df["DATE"].unique())



"""Module to hold Eclipse input and output filenames"""

import errno
import logging
import os
from pathlib import Path
from typing import Any, List, Optional, Tuple, Union

try:
    import opm.io

    HAVE_OPM = True
except ImportError:
    HAVE_OPM = False

from ecl import EclFileFlagEnum
from ecl.eclfile import EclFile
from ecl.grid import EclGrid
from ecl.summary import EclSum

from ecl2df import common

logger = logging.getLogger(__name__)

if HAVE_OPM:
    # Default parse option to opm.io for a very permissive parsing
    OPMIOPARSER_RECOVERY: List[Tuple[str, Any]] = [
        ("PARSE_EXTRA_DATA", opm.io.action.ignore),
        ("PARSE_EXTRA_RECORDS", opm.io.action.ignore),
        ("PARSE_INVALID_KEYWORD_COMBINATION", opm.io.action.ignore),
        ("PARSE_MISSING_DIMS_KEYWORD", opm.io.action.ignore),
        ("PARSE_MISSING_INCLUDE", opm.io.action.ignore),
        ("PARSE_MISSING_SECTIONS", opm.io.action.ignore),
        ("PARSE_RANDOM_SLASH", opm.io.action.ignore),
        ("PARSE_RANDOM_TEXT", opm.io.action.ignore),
        ("PARSE_UNKNOWN_KEYWORD", opm.io.action.ignore),
        ("SUMMARY_UNKNOWN_GROUP", opm.io.action.ignore),
        ("UNSUPPORTED_*", opm.io.action.ignore),
    ]


class EclFiles(object):
    """
    Class for holding an Eclipse deck with result files

    Exists only for convenience, so that loading of
    EclFile/EclSum objects is easy for users, and with
    caching if wanted.

    Various functions that needs some of the Eclipse output
    (or input file) should be able to ask this class, and
    it should be loaded or served from cache.
    """

    def __init__(self, eclbase):
        # eclbase might be a a Posix path object
        eclbase = str(eclbase)

        # Hint about possible wrong filenames:
        if ".DATA" in eclbase and not Path(eclbase).is_file():
            logger.warning("File %s does not exist", eclbase)
            # (this is not an error, because it is possible
            # to obtain summary without the DATA file being present)

        # Strip .DATA or . at end of eclbase:
        eclbase = rreplace(".DATA", "", eclbase)
        eclbase = rreplace(".", "", eclbase)
        self._eclbase = eclbase

        # Set class variables to None
        self._egridfile = None  # Should be EclFile
        self._initfile = None  # Should be EclFile
        self._eclsum = None  # Should be EclSum

        self._egrid = None  # Should be EclGrid

        self._rstfile = None  # EclFile
        self._rftfile = None  # EclFile

        self._deck = None

    def get_path(self) -> Path:
        """Return the full path to the directory with the DATA file"""
        return Path(self._eclbase).absolute().parent

    def get_ecldeck(self) -> "opm.libopmcommon_python.Deck":
        """Return a opm.io deck of the DATA file"""
        if not self._deck:
            if Path(self._eclbase + ".DATA").is_file():
                deckfile = self._eclbase + ".DATA"
            else:
                deckfile = self._eclbase  # Will be any filename
            logger.info("Parsing deck file %s...", deckfile)
            parsecontext = opm.io.ParseContext(OPMIOPARSER_RECOVERY)
            deck = opm.io.Parser().parse(deckfile, parsecontext)
            self._deck = deck
        return self._deck

    @staticmethod
    def str2deck(
        string: str, parsecontext: Optional[List[Tuple[str, Any]]] = None
    ) -> "opm.libopmcommon_python.Deck":
        """Produce a opm.io deck from a string, using permissive
        parsing by default"""
        if parsecontext is None:
            parsecontext = opm.io.ParseContext(OPMIOPARSER_RECOVERY)
        return opm.io.Parser().parse_string(string, parsecontext)

    @staticmethod
    def file2deck(filename: Union[str, Path]) -> "opm.libopmcommon_python.Deck":
        """Try to convert standalone files into opm.io Deck objects"""
        return EclFiles.str2deck(Path(filename).read_text(encoding="utf-8"))

    def get_egrid(self) -> EclGrid:
        """Find and return EGRID file as an EclGrid object"""
        if not self._egrid:
            egridfilename = self._eclbase + ".EGRID"
            if not Path(egridfilename).is_file():
                raise FileNotFoundError(
                    errno.ENOENT, os.strerror(errno.ENOENT), egridfilename
                )
            logger.info("Opening grid data from EGRID file: %s", egridfilename)
            self._egrid = EclGrid(egridfilename)
        return self._egrid

    def get_egridfile(self) -> EclFile:
        """Find and return the EGRID file as a EclFile object

        This gives access to data vectors defined on the grid."""
        if not self._egridfile:
            egridfilename = self._eclbase + ".EGRID"
            if not Path(egridfilename).is_file():
                raise FileNotFoundError(
                    errno.ENOENT, os.strerror(errno.ENOENT), egridfilename
                )
            logger.info("Opening data vectors from EGRID file: %s", egridfilename)
            self._egridfile = EclFile(
                egridfilename, flags=EclFileFlagEnum.ECL_FILE_CLOSE_STREAM
            )

        return self._egridfile

    def get_eclsum(self, include_restart: bool = True) -> EclSum:
        """Find and return the summary file and
        return as EclSum object

        Args:
            include_restart: Sent to libecl for whether restart files
                should be traversed.
        """
        if not self._eclsum:
            smryfilename = self._eclbase + ".UNSMRY"
            if not Path(smryfilename).is_file():
                raise FileNotFoundError(
                    errno.ENOENT, os.strerror(errno.ENOENT), smryfilename
                )
            logger.info("Opening UNSMRY file: %s", smryfilename)
            self._eclsum = EclSum(smryfilename, include_restart=include_restart)
        return self._eclsum

    def get_initfile(self) -> EclFile:
        """Find and return the INIT file as an EclFile object"""
        if not self._initfile:
            initfilename = self._eclbase + ".INIT"
            if not Path(initfilename).is_file():
                raise FileNotFoundError(
                    errno.ENOENT, os.strerror(errno.ENOENT), initfilename
                )
            logger.info("Opening INIT file: %s", initfilename)
            self._initfile = EclFile(
                initfilename, flags=EclFileFlagEnum.ECL_FILE_CLOSE_STREAM
            )
        return self._initfile

    def get_rftfile(self) -> EclFile:
        """Find and return the RFT file as an EclFile object"""
        if not self._rftfile:
            rftfilename = self._eclbase + ".RFT"
            if not Path(rftfilename).is_file():
                raise FileNotFoundError(
                    errno.ENOENT, os.strerror(errno.ENOENT), rftfilename
                )
            logger.info("Opening RFT file: %s", rftfilename)
            self._rftfile = EclFile(
                rftfilename, flags=EclFileFlagEnum.ECL_FILE_CLOSE_STREAM
            )
        return self._rftfile

    def get_rstfile(self) -> EclFile:
        """Find and return the UNRST file as an EclFile object"""
        if not self._rstfile:
            rstfilename = self._eclbase + ".UNRST"
            if not Path(rstfilename).is_file():
                raise FileNotFoundError(
                    errno.ENOENT, os.strerror(errno.ENOENT), rstfilename
                )
            logger.info("Opening RST file: %s", rstfilename)
            self._rstfile = EclFile(
                rstfilename, flags=EclFileFlagEnum.ECL_FILE_CLOSE_STREAM
            )
        return self._rstfile

    def get_rstfilename(self) -> str:
        """Return the inferred name of the UNRST file"""
        return self._eclbase + ".UNRST"

    def get_prtfilename(self) -> str:
        """Return the inferred name of the PRT file"""
        return self._eclbase + ".PRT"

    def close(self) -> None:
        """Close any opened files. Most files are opened though ecl with
        an option to close the stream as possible, leaving not much work
        for this function."""
        self._egridfile = None
        self._initfile = None
        # This is necessary for garbage collection to close the EclSum file:
        self._eclsum = None
        self._rstfile = None
        self._rftfile = None

    def get_zonemap(self, filename=None):
        """Return a dictionary from (int) K layers in the simgrid to strings

        Typical usage is to map from grid layer to zone names.

        The layer filename must currently follow format::

          'ZoneA' 1-4
          'ZoneB' 5-10

        where the single quotes are optional for zones without spaces.
        Write single layer zones as 11-11. NB: ResInsight requires single
        quotes always.

        Args:
            filename (str): Name of file. If relative path, relative to DATA
                file location. If nonexisting file, an empty dict will be
                returned and a warning issued.

        Returns:
            dict, integer keys which are the K layers. Every layer mentioned
                in the interval in the input file is present. Can be empty.
        """
        if not filename:
            filename_defaulted = True
            filename = "zones.lyr"
        else:
            filename_defaulted = False
        assert isinstance(filename, str)
        if not Path(filename).is_absolute():
            fullpath = Path(self.get_path()) / filename
        else:
            fullpath = filename
        if not Path(fullpath).is_file():
            if filename_defaulted:
                # No warnings when the default filename is not there.
                return {}
            logger.warning("Zonefile %s not found, ignoring", fullpath)
            return {}
        lyrlist = common.parse_lyrfile(fullpath)
        return common.convert_lyrlist_to_zonemap(lyrlist)


def rreplace(pat: str, sub: str, string: str) -> str:
    """Variant of str.replace() that only replaces at the end of the string"""
    return string[0 : -len(pat)] + sub if string.endswith(pat) else string  # noqa



"""Exctracts connection status history for each well connections"""

import argparse
import logging
import re
from typing import Any, List, Tuple

import numpy as np
import pandas as pd

from ecl2df import getLogger_ecl2csv, summary
from ecl2df.eclfiles import EclFiles

from .common import write_dframe_stdout_file

logger = logging.getLogger(__name__)


def df(eclfiles: EclFiles) -> pd.DataFrame:
    """Exctracts connection status history for each compdat connection that
    is included in the summary data on the form CPI:WELL,I,J,K. CPI stands for
    connection productivity index.

    One line is added to the export every time a connection changes status. It
    is OPEN when CPI>0 and SHUT when CPI=0. The earliest date for any connection
    will be OPEN, i.e a cell can not be SHUT before it has been OPEN. This means
    that any cells that are always SHUT will be excluded.

    The output data set is very sparse compared to the CPI summary data.
    """
    smry = summary.df(eclfiles, column_keys="CPI*")
    return _extract_status_changes(smry)


def _extract_status_changes(smry: pd.DataFrame) -> pd.DataFrame:
    """Extracts connections status changes from a dataframe of CPI
    summary data.
    """
    cpi_columns = [
        col
        for col in smry.columns
        if re.match("^CPI:[A-Z0-9_-]{1,8}:[0-9]+,[0-9]+,[0-9]+$", col)
    ]
    dframe = pd.DataFrame(columns=["DATE", "WELL", "I", "J", "K", "OP/SH"])

    for col in cpi_columns:
        colsplit = col.split(":")
        well = colsplit[1]
        i, j, k = colsplit[2].split(",")

        status_changes = _extract_single_connection_status_changes(
            smry.index, smry[col]
        )
        for date, status in status_changes:
            dframe.loc[dframe.shape[0]] = [date, well, i, j, k, status]

    logger.info(
        "Dataframe with well connection status ready, %d rows",
        len(dframe),
    )
    return dframe


def _extract_single_connection_status_changes(
    dates: np.ndarray, conn_values: np.ndarray
) -> List[Tuple[Any, str]]:
    """Extracts the status history of a single connection as a list of tuples
    on the form (date, status)

    A CPI value of 0 means that the connection is SHUT
    A CPI value > 0 means that the connection is OPEN
    """
    status_changes = []
    prev_value = 0
    for date, value in zip(dates, conn_values):
        if value > 0 and prev_value == 0:
            # Connection is OPEN and was SHUT at previous timestep
            status_changes.append((date, "OPEN"))
        elif prev_value > 0 and value == 0:
            # Connection is SHUT and was OPEN at previous timestep
            status_changes.append((date, "SHUT"))
        prev_value = value
    return status_changes


def fill_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Set up sys.argv parsers.

    Arguments:
        parser: parser to fill with arguments
    """
    parser.add_argument(
        "DATAFILE",
        type=str,
        help="Name of Eclipse DATA file. " + "UNSMRY file must lie alongside.",
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help=(
            "Name of output csv file. Use '-' to write to stdout. "
            "Default 'well_connection_status.csv'"
        ),
        default="well_connection_status.csv",
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="Be verbose")
    return parser


def wellconnstatus_main(args):
    """Entry-point for module, for command line utility"""
    logger = getLogger_ecl2csv(__name__, vars(args))
    eclfiles = EclFiles(args.DATAFILE)

    wellconnstatus_df = df(eclfiles)
    write_dframe_stdout_file(
        wellconnstatus_df, args.output, index=False, caller_logger=logger
    )



"""Test module for ecl2df.common"""

import os
from pathlib import Path

import pandas as pd
import pytest

from ecl2df import common, eclfiles, equil

try:
    import opm  # noqa

    HAVE_OPM = True
except ImportError:
    HAVE_OPM = False


def test_opmkeywords():
    """Test that we have loaded some keyword metadata from json files on disk"""
    assert "WCONPROD" in common.OPMKEYWORDS
    assert common.OPMKEYWORDS["WCONPROD"]
    assert "name" in common.OPMKEYWORDS["WCONPROD"]
    assert "BHP" in [x["name"] for x in common.OPMKEYWORDS["WCONPROD"]["items"]]

    # This file should not be parsed..
    assert "README" not in common.OPMKEYWORDS


def test_stack_on_colname():
    """Test that we can stack column with an implicit double level
    in the column names indicated by a separator string"""

    dframe = pd.DataFrame(
        columns=["EQLNUM", "STATIC", "OWC@2000-01-01", "OWC@2020-01-01"],
        data=[[1, 1.2, 2000, 1900], [2, 1.3, 2100, 2050]],
    )
    orig_dframe = dframe.copy()
    common.stack_on_colnames(dframe, inplace=False)
    pd.testing.assert_frame_equal(dframe, orig_dframe)

    stacked = common.stack_on_colnames(dframe)
    assert not dframe.equals(orig_dframe)  # It was modifid in the process
    assert "DATE" in stacked
    assert "OWC" in stacked
    assert len(stacked.columns) == 4
    assert len(stacked["DATE"].unique()) == 2
    assert len(stacked) == 4
    assert not stacked.isnull().sum().sum()

    dframe = pd.DataFrame(
        columns=[
            "EQLNUM",
            "STATIC",
            "OWC@2000-01-01",
            "OWC@2020-01-01",
            "GOC@2000-01-01",
            "GOC@2020-01-01",
        ],
        data=[[1, 1.2, 2000, 1900, 1800, 1700], [2, 1.3, 2100, 2050, 2000, 1950]],
    )
    stacked = common.stack_on_colnames(dframe)
    assert "DATE" in stacked
    assert "OWC" in stacked
    assert "GOC" in stacked
    assert len(stacked.columns) == 5
    assert len(stacked["DATE"].unique()) == 2
    assert len(stacked) == 4
    assert not stacked.isnull().sum().sum()

    dframe = pd.DataFrame(
        columns=["OWC@2000-01-01", "OWC@2020-01-01"], data=[[2000, 1900], [2100, 2050]]
    )
    stacked = common.stack_on_colnames(dframe)
    assert "DATE" in stacked
    assert "OWC" in stacked
    assert len(stacked.columns) == 2
    assert len(stacked["DATE"].unique()) == 2
    assert len(stacked) == 4
    assert not stacked.isnull().sum().sum()

    dframe = pd.DataFrame(columns=["EQLNUM", "STATIC"], data=[[1, 1.2], [2, 1.3]])
    stacked = common.stack_on_colnames(dframe)
    assert "DATE" not in stacked
    assert "OWC" not in stacked
    assert "EQLNUM" in stacked
    assert "STATIC" in stacked
    assert len(stacked.columns) == 2
    assert len(stacked) == 2
    assert not stacked.isnull().sum().sum()


def test_write_dframe_file(tmp_path):
    """Test that we can write dataframes to files."""
    os.chdir(tmp_path)
    dframe = pd.DataFrame([{"foo": "bar"}])
    common.write_dframe_stdout_file(dframe, "foo.csv")
    pd.testing.assert_frame_equal(pd.read_csv("foo.csv"), dframe)


def test_write_dframe_stdout(capsys):
    """Test that we can write dataframes to stdout."""
    dframe = pd.DataFrame([{"foo": "bar"}])
    common.write_dframe_stdout_file(dframe, common.MAGIC_STDOUT)
    assert "foo\nbar" in capsys.readouterr().out


def test_write_inc_file(tmp_path):
    """Test that we can write include files to files."""
    os.chdir(tmp_path)
    string = "PORO\n0\n/"
    common.write_inc_stdout_file(string, "poro.inc")
    assert Path("poro.inc").read_text() == string


def test_write_inc_stdout(capsys):
    """Test that we can write include files to stdout."""
    string = "PORO\n0\n/"
    common.write_inc_stdout_file(string, common.MAGIC_STDOUT)
    assert string in capsys.readouterr().out


def test_parse_opmio_deckrecord():
    """Test common properties of the general parse_opmio_deckrecord.

    This function is also indirectly tested in each different
    submodule"""
    with pytest.raises(ValueError, match="Keyword FOOBAR not supported"):
        common.parse_opmio_deckrecord(None, "FOOBAR")


@pytest.mark.skipif(not HAVE_OPM, reason="OPM is not installed")
@pytest.mark.parametrize(
    "wanted, deckstr, supported, expected",
    [
        ("PORO", "PORO\n/", ["PORO"], ["PORO"]),
        (["PORO"], "PORO\n/", ["PORO"], ["PORO"]),
        ("PORO", "PORO\n/", [], []),
        ("PERMX", "PORO\n/", ["PORO"], []),
        ("PERMX", "PERMX\n/", ["PORO"], []),
        ("PERMX", "PORO\n/", ["PERMX"], []),
    ],
)
def test_handle_wanted_keywords(wanted, deckstr, supported, expected):
    """Test that we can handle list of wanted, supported and available keywords."""
    deck = eclfiles.EclFiles.str2deck(deckstr)
    assert common.handle_wanted_keywords(wanted, deck, supported) == expected


def df2ecl_equil(dframe, comment: str = None):
    """Wrapper function to be able to test df2ecl

    (it asks for a function in the calling module)"""
    return equil.df2ecl_equil(dframe, comment)


def test_df2ecl():
    """Test general properties of df2ecl.

    This function is mainly tested in each submodule."""
    dframe = pd.DataFrame(
        [
            {
                "Z": 2469.0,
                "PRESSURE": 382.4,
                "OWC": 100.0,
                "PCOWC": 0.0,
                "GOC": 0.0,
                "EQLNUM": 1,
                "KEYWORD": "EQUIL",
            }
        ]
    )
    with pytest.raises(AssertionError):
        # supported keywords are not supplied
        common.df2ecl(dframe)
    with pytest.raises(AssertionError):
        common.df2ecl(dframe, supported=None)

    with pytest.raises(ValueError, match="KEYWORD must be in the dataframe"):
        common.df2ecl(
            dframe.drop("KEYWORD", axis=1), keywords=["EQUIL"], supported=["EQUIL"]
        )

    string = common.df2ecl(dframe, supported=["EQUIL"])
    # The next calls differ only in timestamp:
    assert len(string) == len(
        common.df2ecl(dframe, keywords="EQUIL", supported=["EQUIL"])
    )
    assert len(string) == len(
        common.df2ecl(dframe, keywords=["EQUIL"], supported=["EQUIL"])
    )
    assert "EQUIL\n" in string
    assert "2469" in string
    assert "-- Output file printed by tests.test_common" in string

    assert "" == common.df2ecl(dframe, supported=["PORO"])

    assert "EQUIL\n-- foobar" in common.df2ecl(
        dframe, comments={"EQUIL": "foobar"}, supported=["EQUIL"]
    )
    assert "\n\n-- masterfoobar\nEQUIL" in common.df2ecl(
        dframe, comments={"master": "masterfoobar"}, supported=["EQUIL"]
    )

    tworows = pd.concat([dframe, dframe])
    tworows["EQLNUM"] = [3, 1]
    tworows["PRESSURE"] = [3456, 1234]
    with pytest.raises(ValueError):
        common.df2ecl(tworows, supported=["EQUIL"], consecutive="EQLNUM")
    # This would be a bug if client code did this, because the wrong
    # consecutive column is set:
    assert "3456" in common.df2ecl(tworows, supported=["EQUIL"], consecutive="PVTNUM")
    tworows["EQLNUM"] = [1, 3]
    with pytest.raises(ValueError):
        common.df2ecl(tworows, supported=["EQUIL"], consecutive="EQLNUM")
    tworows["EQLNUM"] = [2, 1]
    # Passes because the frame is sorted on EQLNUM:
    string = common.df2ecl(tworows, supported=["EQUIL"], consecutive="EQLNUM")
    assert "EQUIL" in string
    assert string.find("3456") > string.find("1234")


def test_eclcompress():
    """Test that we can compress string using Eclipse style
    run-length encoding"""
    assert common.runlength_eclcompress("") == ""
    assert common.runlength_eclcompress(" ") == ""
    assert common.runlength_eclcompress("1 2") == "1  2"
    assert common.runlength_eclcompress("1 2", sep=" ") == "1 2"
    assert common.runlength_eclcompress("1 2", sep="   ") == "1   2"
    assert common.runlength_eclcompress("1") == "1"
    assert common.runlength_eclcompress("1 1") == "2*1"
    assert common.runlength_eclcompress("1 1 1") == "3*1"
    assert common.runlength_eclcompress("1     1 1") == "3*1"
    assert common.runlength_eclcompress("1  \n  1 1 2") == "3*1  2"


@pytest.mark.parametrize(
    "template, wells, output",
    [
        ("OP*", ["OP1", "OP2", "WI"], ["OP1", "OP2"]),
        ("B*H", ["B_1H", "BH", "B_23H", "WI"], ["B_1H", "BH", "B_23H"]),
        ("B_1H*", ["B_1H", "B_1HT2", "OB_1H"], ["B_1H", "B_1HT2"]),
        ("\\*P1", ["OP1", "WI"], ["OP1"]),
        ("B_?H", ["B_1H", "B_12H"], ["B_1H"]),
        ("\\????", ["B_1H", "D_2H", "OP1"], ["B_1H", "D_2H"]),
        pytest.param(
            "*P1",
            ["OP1"],
            None,
            marks=pytest.mark.xfail(
                raises=ValueError,
                match="Well template not allowed to start with a wildcard character",
            ),
        ),
        pytest.param(
            "????",
            ["B_1H"],
            None,
            marks=pytest.mark.xfail(
                raises=ValueError,
                match="Well template not allowed to start with a wildcard character",
            ),
        ),
    ],
)
def test_well_matching_template(template, wells, output):
    "Test that get_wells_matching_template is working as intended."
    assert common.get_wells_matching_template(template, wells) == output



"""Test module for satfunc2df"""

import os
import subprocess
from pathlib import Path

import numpy as np
import pandas as pd
import pytest

from ecl2df import csv2ecl, ecl2csv, inferdims, satfunc
from ecl2df.eclfiles import EclFiles

try:
    import opm  # noqa
except ImportError:
    pytest.skip(
        "OPM is not installed, nothing relevant in here then",
        allow_module_level=True,
    )


TESTDIR = Path(__file__).absolute().parent
REEK = str(TESTDIR / "data/reek/eclipse/model/2_R001_REEK-0.DATA")
EIGHTCELLS = str(TESTDIR / "data/eightcells/EIGHTCELLS.DATA")


def test_ecldeck_to_satfunc_dframe():
    """Test that dataframes can be produced from a full Eclipse deck (the
    example Reek case)"""
    eclfiles = EclFiles(REEK)
    satdf = satfunc.df(eclfiles.get_ecldeck())

    assert set(satdf["KEYWORD"]) == {"SWOF", "SGOF"}
    assert set(satdf["SATNUM"]) == {1}

    assert np.isclose(satdf["SW"].min(), 0.32)
    assert np.isclose(satdf["SW"].max(), 1.0)

    assert np.isclose(satdf["SG"].min(), 0.0)
    assert np.isclose(satdf["SG"].max(), 1 - 0.32)

    assert np.isclose(satdf["KRW"].min(), 0.0)
    assert np.isclose(satdf["KRW"].max(), 1.0)

    assert np.isclose(satdf["KROW"].min(), 0.0)
    assert np.isclose(satdf["KROW"].max(), 1.0)

    assert np.isclose(satdf["KROG"].min(), 0.0)
    assert np.isclose(satdf["KROG"].max(), 1.0)

    assert len(satdf) == 76


def test_satfunc_roundtrip():
    """Test that we can produce a SATNUM dataframe from the Reek case, convert
    it back to an include file, and then reinterpret it to the same"""
    eclfiles = EclFiles(EIGHTCELLS)
    satdf = satfunc.df(eclfiles.get_ecldeck())
    inc = satfunc.df2ecl(satdf)
    df_from_inc = satfunc.df(inc)
    pd.testing.assert_frame_equal(
        satdf.sort_values(["SATNUM", "KEYWORD"]),
        df_from_inc.sort_values(["SATNUM", "KEYWORD"]),
    )


def test_df2ecl_order():
    """Test that we can control the keyword order in generated
    strings by the list supplied in keywords argument"""
    eclfiles = EclFiles(REEK)
    satdf = satfunc.df(eclfiles.get_ecldeck())

    swof_sgof = satfunc.df2ecl(satdf, keywords=["SWOF", "SGOF"])
    assert swof_sgof.find("SWOF") < swof_sgof.find("SGOF")
    sgof_swof = satfunc.df2ecl(satdf, keywords=["SGOF", "SWOF"])
    assert sgof_swof.find("SGOF") < sgof_swof.find("SWOF")

    only_swof = satfunc.df2ecl(satdf, keywords=["SWOF"])
    assert "SGOF" not in only_swof
    only_sgof = satfunc.df2ecl(satdf, keywords="SGOF")
    assert "SWOF" not in only_sgof


def test_nodata():
    """Test when no data is found"""
    swofstr = ""

    satdf = satfunc.df(swofstr)
    assert len(satdf) == 0

    inc = satfunc.df2ecl_swof(satdf)
    assert "No data" in inc
    df_from_inc = satfunc.df(inc)
    assert df_from_inc.empty


@pytest.mark.parametrize(
    "string, expected_df",
    [
        ("", pd.DataFrame()),
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        (
            """SWOF
  0 0 1 1
  1 1 0 0
/
""",
            pd.DataFrame(
                columns=["SW", "KRW", "KROW", "PCOW", "SATNUM", "KEYWORD"],
                data=[[0.0, 0.0, 1.0, 1.0, 1, "SWOF"], [1.0, 1.0, 0.0, 0.0, 1, "SWOF"]],
            ),
        ),
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        (
            """
RUNSPEC -- (this line is optional)

TABDIMS
  2/

PROPS -- (optional)

SWOF
  0 0 1 1
  1 1 0 0
/
  0 0 1 1
  0.5 0.5 0.5 0.5
  1 1 0 0
/
""",
            pd.DataFrame(
                columns=["SW", "KRW", "KROW", "PCOW", "SATNUM", "KEYWORD"],
                data=[
                    [0.0, 0.0, 1.0, 1.0, 1, "SWOF"],
                    [1.0, 1.0, 0.0, 0.0, 1, "SWOF"],
                    [0.0, 0.0, 1.0, 1.0, 2, "SWOF"],
                    [0.5, 0.5, 0.5, 0.5, 2, "SWOF"],
                    [1.0, 1.0, 0.0, 0.0, 2, "SWOF"],
                ],
            ),
        ),
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        ("SWRF\n0 / \n", pd.DataFrame()),  # a warning will be printed
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        (
            # Bogus E100 keywords, will be accepted:
            "FOO\n\nSWOF\n 0 0 1 1\n 1 1  0 0\n/\n",
            pd.DataFrame(
                columns=["SW", "KRW", "KROW", "PCOW", "SATNUM", "KEYWORD"],
                data=[[0.0, 0.0, 1.0, 1.0, 1, "SWOF"], [1.0, 1.0, 0.0, 0.0, 1, "SWOF"]],
            ),
        ),
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        (
            """SLGOF
  0 1 1 0
  1 0 0 0
/
    """,
            pd.DataFrame(
                columns=["SL", "KRG", "KRO", "PCOG", "SATNUM", "KEYWORD"],
                data=[
                    [0.0, 1.0, 1.0, 0.0, 1, "SLGOF"],
                    [1.0, 0.0, 0.0, 0.0, 1, "SLGOF"],
                ],
            ),
        ),
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        (
            """SOF2
  0 1
  1 0
/
""",
            pd.DataFrame(
                columns=["SO", "KRO", "SATNUM", "KEYWORD"],
                data=[
                    [0.0, 1.0, 1, "SOF2"],
                    [1.0, 0.0, 1, "SOF2"],
                ],
            ),
        ),
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        (
            """SOF3
  0 1 1
  1 0 0
/
""",
            pd.DataFrame(
                columns=["SO", "KROW", "KROG", "SATNUM", "KEYWORD"],
                data=[
                    [0.0, 1.0, 1.0, 1, "SOF3"],
                    [1.0, 0.0, 0.0, 1, "SOF3"],
                ],
            ),
        ),
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        (
            """
SGFN
  0 1 0
  1 0 0
/
  0 1 0
  1 0.1 1
/
""",
            pd.DataFrame(
                columns=["SG", "KRG", "PCOG", "SATNUM", "KEYWORD"],
                data=[
                    [0.0, 1.0, 0.0, 1, "SGFN"],
                    [1.0, 0.0, 0.0, 1, "SGFN"],
                    [0.0, 1.0, 0.0, 2, "SGFN"],
                    [1.0, 0.1, 1.0, 2, "SGFN"],
                ],
            ),
        ),
        # # # # # # # # # # # # # # # # # # # # # # # # # # # # # #
        (
            """SGWFN
  0 1 1 0
  1 0 0 0
 /
 """,
            pd.DataFrame(
                columns=["SG", "KRG", "KRW", "PCGW", "SATNUM", "KEYWORD"],
                data=[
                    [0.0, 1.0, 1.0, 0.0, 1, "SGWFN"],
                    [1.0, 0.0, 0.0, 0.0, 1, "SGWFN"],
                ],
            ),
        ),
    ],
)
def test_str2df(string, expected_df):
    """Test that we can parse a string into a DataFrame,
    back to string, and to DataFrame again"""
    satdf = satfunc.df(string)
    pd.testing.assert_frame_equal(satdf, expected_df)

    if expected_df.empty:
        return

    inc = satfunc.df2ecl(satdf)
    df_from_inc = satfunc.df(inc)
    pd.testing.assert_frame_equal(df_from_inc, expected_df)


def test_sgof_satnuminferrer(tmp_path, mocker):
    """Test inferring of SATNUMS in SGOF strings"""
    sgofstr = """
SGOF
  0 0 1 1
  1 1 0 0
/
  0 0 1 1
  0.5 0.5 0.5 0.5
  1 1 0 0
/
  0 0 1 0
  0.1 0.1 0.1 0.1
  1 1 0 0
/
"""
    os.chdir(tmp_path)
    assert inferdims.guess_dim(sgofstr, "TABDIMS", 0) == 3
    sgofdf = satfunc.df(sgofstr)
    assert "SATNUM" in sgofdf
    assert len(sgofdf["SATNUM"].unique()) == 3
    assert len(sgofdf) == 8
    inc = satfunc.df2ecl(sgofdf)
    df_from_inc = satfunc.df(inc)
    pd.testing.assert_frame_equal(sgofdf, df_from_inc)

    # Write to file and try to parse it with command line:
    sgoffile = "__sgof_tmp.txt"
    Path(sgoffile).write_text(sgofstr)
    mocker.patch(
        "sys.argv", ["ecl2csv", "satfunc", "-v", sgoffile, "-o", sgoffile + ".csv"]
    )
    ecl2csv.main()
    parsed_sgof = pd.read_csv(sgoffile + ".csv")
    assert len(parsed_sgof["SATNUM"].unique()) == 3


def test_wrong_columns():
    """Test some error situations"""
    # SWFN data given as SWOF:
    satnumstr = """
SWOF
0 0 0
1 1 0
/
"""
    with pytest.raises(ValueError, match="Wrong number count for keyword SWOF"):
        satfunc.df(satnumstr)
    satnumstr = """
SWFN
0 0 0 0
1 1 0 0
/
"""
    with pytest.raises(ValueError, match="Wrong number count for keyword SWFN"):
        satfunc.df(satnumstr)

    # The following error is parseable into a dataframe, but gives
    # four saturation points, this error can not be detected while parsing.
    satnumstr = """
SWFN
0 0 0 0
0.5 0.5 0.5 0
1 1 0 0
/
"""
    wrongdf = satfunc.df(satnumstr)
    # We see the error as the saturation points are not unique:
    assert len(wrongdf["SW"]) == 4
    assert len(wrongdf["SW"].unique()) == 3


def test_defaulted_swof_values():
    """The Eclipse manual states that missing values in
    a SWOF/SWFN/++ record should be replaced by linearly interpolated values"""
    dframe = satfunc.df(
        """
SWOF
0 0 1 1
0.5 1* 1* 1*
1 1 0 0
/
"""
    )
    np.testing.assert_allclose(dframe["PCOW"], [1, 0.5, 0])
    np.testing.assert_allclose(dframe["KROW"], [1, 0.5, 0])
    np.testing.assert_allclose(dframe["KRW"], [0, 0.5, 1])


def test_defaulted_sgof_values():
    """The Eclipse manual states that missing values in
    a SWOF/SWFN/++ record should be replaced by linearly interpolated values"""
    dframe = satfunc.df(
        """
SGOF
0 0 1 1
0.5 1* 1* 1*
1 1 0 0
/
0 0 1 1
0.2 1* 1* 1*
1 1 0 0
/
"""
    )
    np.testing.assert_allclose(dframe["PCOG"], [1, 0.5, 0, 1, 0.8, 0])
    np.testing.assert_allclose(dframe["KROG"], [1, 0.5, 0, 1, 0.8, 0])
    np.testing.assert_allclose(dframe["KRG"], [0, 0.5, 1, 0, 0.2, 1])


def test_defaulted_swfn_values():
    """The Eclipse manual states that missing values in
    a SWOF/SWFN/++ record should be replaced by linearly interpolated values"""
    dframe = satfunc.df(
        """
SWFN
0 0 1
0.1 1* 1*
0.9 1* 1*
1 1 0
/
"""
    )
    np.testing.assert_allclose(dframe["PCOW"], [1, 0.9, 0.1, 0])
    np.testing.assert_allclose(dframe["KRW"], [0, 0.1, 0.9, 1])


def test_defaulted_sof3_values():
    """The Eclipse manual states that missing values in
    a SWOF/SWFN/++ record should be replaced by linearly interpolated values"""
    dframe = satfunc.df(
        """
SOF3
0 0 1
0.1 1* 1*
0.9 1* 1*
1 1 0
/
"""
    )
    np.testing.assert_allclose(dframe["KROG"], [1, 0.9, 0.1, 0])
    np.testing.assert_allclose(dframe["KROW"], [0, 0.1, 0.9, 1])


def test_defaulted_sof2_values():
    """The Eclipse manual states that missing values in
    a SWOF/SWFN/++ record should be replaced by linearly interpolated values"""
    dframe = satfunc.df(
        """
SOF2
0 0
0.1 1*
0.9 1*
1 1
/
"""
    )
    np.testing.assert_allclose(dframe["KRO"], [0, 0.1, 0.9, 1])


def test_defaulted_sgfn_values():
    """The Eclipse manual states that missing values in
    a SWOF/SWFN/++ record should be replaced by linearly interpolated values"""
    dframe = satfunc.df(
        """
SGFN
0 0 1
0.1 1* 1*
0.9 2*
1 1 0
/
"""
    )
    np.testing.assert_allclose(dframe["PCOG"], [1, 0.9, 0.1, 0])
    np.testing.assert_allclose(dframe["KRG"], [0, 0.1, 0.9, 1])


def test_defaulted_sgwfn_values():
    """The Eclipse manual states that missing values in
    a SWOF/SWFN/++ record should be replaced by linearly interpolated values"""
    dframe = satfunc.df(
        """
SGWFN
0 0 1 1
0.5 3*
1 1 0 0
/
"""
    )
    np.testing.assert_allclose(dframe["PCGW"], [1, 0.5, 0])
    np.testing.assert_allclose(dframe["KRG"], [0, 0.5, 1])
    np.testing.assert_allclose(dframe["KRW"], [1, 0.5, 0])


def test_defaulted_slgof_values():
    """The Eclipse manual states that missing values in
    a SWOF/SWFN/++ record should be replaced by linearly interpolated values"""
    dframe = satfunc.df(
        """
SLGOF
0 0 1 1
0.5 3*
1 1 0 0
/
"""
    )
    np.testing.assert_allclose(dframe["PCOG"], [1, 0.5, 0])
    np.testing.assert_allclose(dframe["KRG"], [0, 0.5, 1])
    np.testing.assert_allclose(dframe["KRO"], [1, 0.5, 0])

    # This will probably crash Eclipse:
    dframe = satfunc.df(
        """
SWFN
0 0 1*
1 1 0
/
"""
    )
    np.testing.assert_allclose(dframe["PCOW"], [np.nan, 0])
    np.testing.assert_allclose(dframe["KRW"], [0, 1])


@pytest.mark.parametrize(
    "dframe, expected",
    [
        pytest.param(
            pd.DataFrame([{"SW": 0, "KR": 0, "SATNUM": 1}]),
            pd.DataFrame([{"SW": 0, "KR": 0, "SATNUM": 1}]),
            id="Dummy one-line input, not touched",
        ),
        pytest.param(
            pd.DataFrame(
                [
                    {"SW": 0, "KR": 0, "SATNUM": 1},
                    {"SW": 0.5, "SATNUM": 1},
                    {"SW": 1, "KR": 1, "SATNUM": 1},
                ]
            ),
            pd.DataFrame(
                [
                    {"SW": 0, "KR": 0, "SATNUM": 1},
                    {"SW": 0.5, "KR": 0.5, "SATNUM": 1},
                    {"SW": 1, "KR": 1, "SATNUM": 1},
                ]
            ),
            id="Implicit Nan through Pandas dataframe initialization",
        ),
        pytest.param(
            pd.DataFrame(
                [
                    {"SW": 0, "KR": 0, "SATNUM": 1},
                    {"SW": 0.1, "SATNUM": 1},
                    {"SW": 1, "KR": 1, "SATNUM": 1},
                ]
            ),
            pd.DataFrame(
                [
                    {"SW": 0, "KR": 0, "SATNUM": 1},
                    {"SW": 0.1, "KR": 0.1, "SATNUM": 1},
                    {"SW": 1, "KR": 1, "SATNUM": 1},
                ]
            ),
            id="Ensure linear interpolation in index values",
        ),
        pytest.param(
            pd.DataFrame(
                [
                    {"SW": 0, "KR": 0, "SATNUM": 1},
                    {"SW": 0.1, "KR": np.nan, "SATNUM": 1},
                    {"SW": 1, "KR": 1, "SATNUM": 1},
                ]
            ),
            pd.DataFrame(
                [
                    {"SW": 0, "KR": 0, "SATNUM": 1},
                    {"SW": 0.1, "KR": 0.1, "SATNUM": 1},
                    {"SW": 1, "KR": 1, "SATNUM": 1},
                ]
            ),
            id="Explicit numpy nan",
        ),
        pytest.param(
            pd.DataFrame(
                [
                    {"SW": 0, "KR": 0, "SATNUM": 1},
                    {"SW": 0.1, "KR": float("nan"), "SATNUM": 1},
                    {"SW": 1, "KR": 1, "SATNUM": 1},
                ]
            ),
            pd.DataFrame(
                [
                    {"SW": 0, "KR": 0, "SATNUM": 1},
                    {"SW": 0.1, "KR": 0.1, "SATNUM": 1},
                    {"SW": 1, "KR": 1, "SATNUM": 1},
                ]
            ),
            id="Python float nan",
        ),
        pytest.param(
            pd.DataFrame(
                [
                    {"SW": 0, "FOO": 0, "SATNUM": 1},
                    {"SW": 0.1, "SATNUM": 1},
                    {"SW": 1, "FOO": 1, "SATNUM": 1},
                ]
            ),
            pd.DataFrame(
                [
                    {"SW": 0, "FOO": 0, "SATNUM": 1},
                    {"SW": 0.1, "FOO": 0.1, "SATNUM": 1},
                    {"SW": 1, "FOO": 1, "SATNUM": 1},
                ]
            ),
            id="Will not care about column names",
        ),
        pytest.param(
            pd.DataFrame(
                [
                    {"SW": 0, "KRW": 0, "SATNUM": 1},
                    {"SW": 0.1, "KRW": np.nan, "SATNUM": 1},
                ]
            ),
            pd.DataFrame(
                [
                    {"SW": 0, "KRW": 0, "SATNUM": 1},
                    {"SW": 0.1, "KRW": np.nan, "SATNUM": 1},
                ]
            ),
            id="No extrapolation",
        ),
        pytest.param(
            pd.DataFrame([{"SW": 0}]),
            None,
            marks=pytest.mark.xfail(
                raises=KeyError,
                match="SATNUM",
            ),
        ),
        pytest.param(
            pd.DataFrame([{"SW": np.nan, "SATNUM": 1}]),
            None,
            marks=pytest.mark.xfail(
                raises=ValueError, match="nan in saturation column is not allowed"
            ),
        ),
        pytest.param(
            pd.DataFrame(),
            None,
            marks=pytest.mark.xfail(
                raises=AssertionError,
                match="Could not determine a single saturation column",
            ),
        ),
        pytest.param(
            pd.DataFrame([{"FOO": 1}]),
            None,
            marks=pytest.mark.xfail(
                raises=AssertionError,
                match="Could not determine a single saturation column",
            ),
        ),
        pytest.param(
            pd.DataFrame([{"SW": 1, "SG": 1}]),
            None,
            marks=pytest.mark.xfail(
                raises=AssertionError,
                match="Could not determine a single saturation column",
            ),
        ),
    ],
)
def test_interpolate_defaults(dframe, expected):
    """Test that NaN's in non-saturation columns will be interpolated linearly"""
    pd.testing.assert_frame_equal(satfunc.interpolate_defaults(dframe), expected)


def test_multiple_keywords_family2():

    satnumstr = """
SWFN
-- Sw           Krw           Pcow
  0 0 2
  1.   1.000   0.00000e+00
/

SOF3
-- So           Krow          Krog
   0.00000e+00   0.00000e+00   0.00000e+00
   0.581051658   1.000000000   1.000000000
/

SGFN
-- Sg    Krg      Pcog
  0.000  0.00000  0.000
  0.800  1.00000  0.000
/
    """
    satnum_df = satfunc.df(satnumstr)
    assert set(satnum_df["SATNUM"]) == {1}
    assert set(satnum_df["KEYWORD"]) == {"SWFN", "SOF3", "SGFN"}
    assert len(satnum_df) == 6


def test_main_subparsers(tmp_path, mocker):
    """Test command line interface"""
    tmpcsvfile = tmp_path / "satfunc.csv"
    mocker.patch("sys.argv", ["ecl2csv", "satfunc", EIGHTCELLS, "-o", str(tmpcsvfile)])
    ecl2csv.main()

    assert Path(tmpcsvfile).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile))
    assert not disk_df.empty

    tmpcsvfile2 = tmp_path / ".TMP-satfunc-swof.csv"
    print(tmpcsvfile2)
    mocker.patch(
        "sys.argv",
        [
            "ecl2csv",
            "satfunc",
            EIGHTCELLS,
            "--keywords",
            "SWOF",
            "--output",
            str(tmpcsvfile2),
        ],
    )
    ecl2csv.main()

    assert Path(tmpcsvfile2).is_file()
    disk_df = pd.read_csv(str(tmpcsvfile2))
    assert set(disk_df["KEYWORD"].unique()) == {"SWOF"}


def test_csv2ecl(tmp_path, mocker):
    """Test command line interface for csv to Eclipse include files"""
    os.chdir(tmp_path)
    tmpcsvfile = "satfunc.csv"

    swof_df = pd.DataFrame(
        columns=["KEYWORD", "SW", "KRW", "KROW", "PCOW"],
        data=[["SWOF", 0.0, 0.0, 1.0, 0.0], ["SWOF", 1.0, 1.0, 0.0, 0.0]],
    )
    swof_df.to_csv(tmpcsvfile, index=False)
    mocker.patch("sys.argv", ["csv2ecl", "satfunc", "--output", "swof.inc", tmpcsvfile])
    csv2ecl.main()
    pd.testing.assert_frame_equal(
        satfunc.df(open("swof.inc").read()).drop("SATNUM", axis="columns"),
        swof_df,
        check_like=True,
    )

    # Test writing to stdout:
    result = subprocess.run(
        ["csv2ecl", "satfunc", "--output", "-", tmpcsvfile], stdout=subprocess.PIPE
    )
    pd.testing.assert_frame_equal(
        satfunc.df(result.stdout.decode()).drop("SATNUM", axis="columns"),
        swof_df,
        check_like=True,
    )



import importlib
from pathlib import Path

from pkg_resources import resource_filename

try:
    from ert_shared.plugins.plugin_manager import hook_implementation
    from ert_shared.plugins.plugin_response import plugin_response
except ModuleNotFoundError:
    # ert is not installed - use dummy/transparent function decorators.
    def hook_implementation(func):
        return func

    def plugin_response(plugin_name):  # pylint: disable=unused-argument
        def decorator(func):
            return func

        return decorator


def _get_jobs_from_directory(directory):
    resource_directory = Path(resource_filename("ecl2df", directory))

    all_files = [
        resource_directory / filename
        for filename in resource_directory.glob("*")
        if (resource_directory / filename).exists()
    ]
    return {path.name: str(path) for path in all_files}


@hook_implementation
@plugin_response(plugin_name="ecl2df")
def installable_jobs():
    return _get_jobs_from_directory("config_jobs")


def _get_module_variable_if_exists(module_name, variable_name, default=""):
    try:
        script_module = importlib.import_module(module_name)
    except ImportError:
        return default

    return getattr(script_module, variable_name, default)


@hook_implementation
@plugin_response(plugin_name="ecl2df")
def job_documentation(job_name):
    ecl2df_jobs = set(installable_jobs().data.keys())
    if job_name not in ecl2df_jobs:
        return None

    module_name = f"ecl2df.{job_name.lower()}"

    description = _get_module_variable_if_exists(
        module_name=module_name, variable_name="DESCRIPTION"
    )
    examples = _get_module_variable_if_exists(
        module_name=module_name, variable_name="EXAMPLES"
    )
    category = _get_module_variable_if_exists(
        module_name=module_name, variable_name="CATEGORY", default="other"
    )

    return {
        "description": description,
        "examples": examples,
        "category": category,
    }



#!/usr/bin/env python
"""
End-user command line tool for accessing functionality
in ecl2df
"""
import argparse
import functools
import importlib
import sys
from typing import Optional

from ecl2df import __version__

# String constants in use for generating ERT forward model documentation:
DESCRIPTION: str = """Convert Eclipse input and output files into CSV files,
with the command line utility ``ecl2csv``. Run ``ecl2csv --help`` to see
which subcommands are supported.

For supplying options to subcommands, you can use the arguments ``<XARGn>``
where ``n`` goes from 1 to 10.

For more documentation, see https://equinor.github.io/ecl2df/.
"""
CATEGORY: str = "utility.eclipse"
EXAMPLES: str = """

Outputting the EQUIL data from an Eclipse deck. The ECLBASE variable from your
ERT config is supplied implicitly::

   FORWARD_MODEL ECL2CSV(<SUBCOMMAND>=equil, <OUTPUT>=equil.csv)

For a yearly summary export of the realization, options have to be supplied
with the XARG options::

  FORWARD_MODEL ECL2CSV(<SUBCOMMAND>=summary, <OUTPUT>=yearly.csv, <XARG1>="--time_index", <XARG2>="yearly")

The quotes around double-dashed options are critical to avoid ERT taking for a
comment. For more options, use ``<XARG3>`` etc.
"""  # noqa


def get_parser() -> argparse.ArgumentParser:
    """Make parser"""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=(
            "ecl2csv (" + __version__ + ") is a command line frontend to ecl2df. "
            "Documentation at https://equinor.github.io/ecl2df/ "
        ),
    )
    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )

    if sys.version_info.major >= 3 and sys.version_info.minor >= 7:
        subparsers = parser.add_subparsers(  # type: ignore
            required=True,
            dest="subcommand",
            parser_class=argparse.ArgumentParser,
        )
    else:
        subparsers = parser.add_subparsers(parser_class=argparse.ArgumentParser)

    subparsers_dict = {}
    subparsers_dict["grid"] = subparsers.add_parser(
        "grid",
        help=("Extract grid data with properties"),
        description=(
            "Each cell is represented by "
            "one row of data. The coordinates are in the X, Y, and Z columns "
            "and represent the grid centre. Volume pr. cell is added "
            "and all INIT and UNRST data can be added to the rows"
        ),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    subparsers_dict["summary"] = subparsers.add_parser(
        "summary",
        help=("Extract summary data"),
        description=(
            "This is the time-dependent data for "
            "field production data, well profiles etc. Each row contains data "
            "for one point in time"
        ),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    subparsers_dict["nnc"] = subparsers.add_parser(
        "nnc",
        help="Extract NNC data from EGRID file",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=(
            "Extract NNC (Non-Neighbour Connection) data from the EGRID file. "
            "Each row is one connection, with the columns I1, J1, K1 for the first "
            "cell in the cell pair, and I2, J2, K2 for the second. The "
            "transmissibility for the cell pair is in the TRAN column. "
            "See also the trans subcommand."
        ),
    )
    subparsers_dict["faults"] = subparsers.add_parser(
        "faults",
        help="Extract data from the FAULTS keyword",
        description=(
            "Each row represents a particular cell and a face and the name of the fault"
        ),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    subparsers_dict["trans"] = subparsers.add_parser(
        "trans",
        help="Extract transmissibilities from EGRID file",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=(
            "Extract transmissibilities (TRANX, TRANY, TRANZ) from Eclipse "
            "binary output files. Each row represent a connection between a cell pair "
            "(I1, J1, K1) and (I2, J2, K2). It is possible to add INIT vectors for "
            "each of the cell in the cell pair, e.g. FIPNUM can be added as FIPNUM1 "
            "and FIPNUM2, and it is possible to filter to connections where f.ex "
            "FIPNUM change"
        ),
    )
    subparsers_dict["pillars"] = subparsers.add_parser(
        "pillars",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        help="Compute data pr. cornerpoint pillar",
        description=(
            "Compute statistics pr. pillar in a cornerpoint grid, "
            "or alternatively by region parameter. Volumetrics, in-place, "
            "and contacts."
        ),
    )
    subparsers_dict["pvt"] = subparsers.add_parser(
        "pvt",
        help="Extract PVT data",
        description=(
            "Extract data for the PVT keywords in an Eclipse deck "
            "and merge all data into a single dataframe. "
            "Supported keywords are PVTO, PVDO, PVTG, PVDG, PVTW, "
            "ROCK and DENSITY. Gas phase pressure and oil phase "
            "pressure are both called PRESSURE in the resulting "
            "dataframe, similar for volume factors and viscosity. "
            "Deduce meaning for column names from the Eclipse manual. "
            "The column KEYWORD denotes which PVT keyword a particular "
            "data row stems from. "
        ),
    )
    subparsers_dict["rft"] = subparsers.add_parser(
        "rft",
        help=("Extract RFT data from Eclipse binary output files."),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=(
            "Extract RFT data from Eclipse binary output files to CSV. "
            "Each row in the resulting table represents one point in a "
            "particular well at a particular time. "
            "If multisegment wells are found, associated data "
            "to a connection is merged onto the same row as additional columns. "
            "You need the Eclipse keyword WRFTPLT present in your DATA-file to get "
            "the data outputted."
        ),
    )
    subparsers_dict["fipreports"] = subparsers.add_parser(
        "fipreports",
        help=("Extract FIPxxxxx REPORT REGION data from Eclipse PRT output file."),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=(
            "Extract FIPxxxxx REPORT REGION data from PRT file. "
            "This parses currently in-place, outflows to wells and regions, and "
            "material balance errors"
        ),
    )
    subparsers_dict["satfunc"] = subparsers.add_parser(
        "satfunc",
        help="Extract SWOF/SGOF/etc data",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=(
            "Data for all saturation functions are merged "
            "into one dataframe for all SATNUMs. Each row has data for a "
            "saturation point. For SWOF data, all columns related to SGOF "
            "are empty and vice versa"
        ),
    )
    subparsers_dict["fipreports"] = subparsers.add_parser(
        "fipreports",
        help=("Extract FIPxxxxx REPORT REGION data from Eclipse PRT output file."),
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=(
            "Extract FIPxxxxx REPORT REGION data from PRT file. "
            "This parses currently in-place, outflows to wells and regions, and "
            "material balance errors"
        ),
    )
    subparsers_dict["compdat"] = subparsers.add_parser(
        "compdat",
        help="Extract COMPDAT data",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=(
            "Each row represents a cell connection to a well. "
            "Only COMPDAT data is exposed in CSV output currently."
        ),
    )
    subparsers_dict["equil"] = subparsers.add_parser(
        "equil",
        help="Extract EQUIL data",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=("Each row contains the equilibriation data for one EQLNUM."),
    )
    subparsers_dict["gruptree"] = subparsers.add_parser(
        "gruptree",
        help="Extract GRUPTREE data",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=("Each row represents an edge in the GRUPTREE at a specific date."),
    )
    subparsers_dict["wellconnstatus"] = subparsers.add_parser(
        "wellconnstatus",
        help="Extract well connection status",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=(
            "Each row represents an event where a well connection is changing status."
        ),
    )
    subparsers_dict["wcon"] = subparsers.add_parser(
        "wcon",
        help="Extract well control data",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=(
            "Each row represents the control data for a certain "
            "well or well wildcard at a specific date"
        ),
    )

    for submodule, subparser in subparsers_dict.items():
        # Use the submodule's fill_parser() to add the submodule specific
        # arguments:
        importlib.import_module("ecl2df." + submodule).fill_parser(  # type: ignore
            subparser
        )

        # Add empty placeholders, this looks strange but is needed for the
        # ERT forward model frontend, where non-used options must be supplied
        # as empty string arguments (which we should ignore)
        subparser.add_argument(
            "hiddenemptyplaceholders", nargs="*", help=argparse.SUPPRESS
        )

        # Tell argparse which main() function to use for each
        # subparser/submodule:
        subparser.set_defaults(
            func=functools.partial(
                run_subparser_main, submodule=submodule, parser=subparser
            )
        )

    return parser


def run_subparser_main(
    args,
    submodule: str,
    parser: Optional[argparse.ArgumentParser] = None,
) -> None:
    """Wrapper for running the subparsers main() function, with
    custom argument handling.

    In order to support ERT forward model syntax, empty positional
    arguments must be allowed, but ignored. This function takes
    care of shuffling the single required positional argument into
    args.DATAFILE, and will error if more than one positional
    argument is supplied

    This function is to be supplied to argsparse's set_default()
    function, by use of functools.partial so that the submodule_name
    argument can be set.

    Args:
        args (Namespace): argparse argument namespace
        submodule: One of ecl2df's submodules. That module
            must have a function called <submodule>_main()
        parser: Used for raising errors.
    """
    if "DATAFILE" in args:
        positionals = list(filter(len, [args.DATAFILE] + args.hiddenemptyplaceholders))
        args.DATAFILE = "".join([args.DATAFILE] + args.hiddenemptyplaceholders)
    elif "PRTFILE" in args:
        # Special treatment for the fipreports submodule
        positionals = list(filter(len, [args.PRTFILE] + args.hiddenemptyplaceholders))
        args.PRTFILE = "".join([args.PRTFILE] + args.hiddenemptyplaceholders)
    if len(positionals) > 1 and parser is not None:
        parser.error(f"Unknown argument in {positionals}")

    mod = importlib.import_module("ecl2df." + submodule)

    main_func = getattr(mod, submodule + "_main")
    main_func(args)


def main() -> None:
    """Entry point"""
    parser = get_parser()
    args = parser.parse_args()
    if "arrow" in parser.prog:
        args.__dict__["arrow"] = True
    args.func(args)


if __name__ == "__main__":
    main()



import os
import subprocess
from pathlib import Path

import pandas as pd
import pytest

import ecl2df
from ecl2df.hook_implementations import jobs

try:
    # pylint: disable=unused-import

    import ert_shared  # noqa

    HAVE_ERT = True
except ImportError:
    HAVE_ERT = False

TESTDIR = Path(__file__).absolute().parent
DATADIR = TESTDIR / "data/reek/eclipse/model"


@pytest.mark.skipif(
    not HAVE_ERT, reason="ERT is not installed, skipping hook implementation tests."
)
def test_ecl2csv_through_ert(tmp_path):
    """Test running the ERT executable on a mocked config file"""
    os.chdir(tmp_path)

    # Symlink Eclipse output to our tmp_path:
    eclbase = "2_R001_REEK-0"
    ecl_extensions = [
        "DATA",
        "ECLEND",
        "EGRID",
        "INIT",
        "PRT",
        "RFT",
        "SMSPEC",
        "UNRST",
        "UNSMRY",
    ]

    for ext in ecl_extensions:
        f_name = eclbase + "." + ext
        Path(f_name).symlink_to(DATADIR / f_name)

    ert_config = [
        "ECLBASE " + eclbase + ".DATA",
        "QUEUE_SYSTEM LOCAL",
        "NUM_REALIZATIONS 1",
        "RUNPATH .",
    ]

    csv2ecl_subcommands = ["equil", "pvt", "satfunc"]

    for subcommand in ecl2df.SUBMODULES:
        ert_config.append(
            "FORWARD_MODEL ECL2CSV(<SUBCOMMAND>={0}, <OUTPUT>={0}.csv)".format(
                subcommand
            )
        )

    # Test what we can also supply additional options for some submodules:
    ert_config.append(
        "FORWARD_MODEL ECL2CSV(<SUBCOMMAND>=summary, "
        '<OUTPUT>=summary-yearly.csv, <XARG1>="--time_index", <XARG2>=yearly)'
    )
    ert_config.append(
        "FORWARD_MODEL ECL2CSV(<SUBCOMMAND>=equil, "
        '<OUTPUT>=equil-rsvd.csv, <XARG1>="--keywords", <XARG2>="RSVD")'
    )
    ert_config.append(
        "FORWARD_MODEL ECL2CSV(<SUBCOMMAND>=pvt, "
        '<OUTPUT>=pvt-custom.csv, <XARG1>="--keywords", <XARG2>="PVTO")'
    )
    ert_config.append(
        "FORWARD_MODEL ECL2CSV(<SUBCOMMAND>=satfunc, "
        '<OUTPUT>=satfunc-swof.csv, <XARG1>="--keywords", <XARG2>="SWOF")'
    )

    for subcommand in csv2ecl_subcommands:
        ert_config.append(
            "FORWARD_MODEL CSV2ECL("
            + "<SUBCOMMAND>={0}, <CSVFILE>={0}.csv, <OUTPUT>={0}.inc".format(subcommand)
            + ")"
        )
    ert_config.append(
        "FORWARD_MODEL CSV2ECL(<SUBCOMMAND>=summary, <CSVFILE>=summary-yearly.csv), "
        "<OUTPUT>=SUMYEARLY)"
    )

    ert_config_filename = "ecl2csv_test.ert"
    Path(ert_config_filename).write_text("\n".join(ert_config), encoding="utf-8")

    subprocess.call(["ert", "test_run", ert_config_filename])

    assert Path("OK").is_file()

    for subcommand in ecl2df.SUBMODULES:
        assert Path(subcommand + ".csv").is_file()

    # Check the custom output where options were supplied to the subcommands:
    assert len(pd.read_csv("summary-yearly.csv")) == 5
    assert set(pd.read_csv("equil-rsvd.csv")["KEYWORD"]) == set(["RSVD"])
    assert set(pd.read_csv("pvt-custom.csv")["KEYWORD"]) == set(["PVTO"])
    assert set(pd.read_csv("satfunc-swof.csv")["KEYWORD"]) == set(["SWOF"])

    for subcommand in csv2ecl_subcommands:
        assert Path(subcommand + ".inc").is_file()


@pytest.mark.skipif(not HAVE_ERT, reason="ERT is not installed")
def test_job_documentation():
    if HAVE_ERT:
        assert (
            type(jobs.job_documentation("ECL2CSV"))
            == ert_shared.plugins.plugin_response.PluginResponse
        )
        assert (
            type(jobs.job_documentation("CSV2ECL"))
            == ert_shared.plugins.plugin_response.PluginResponse
        )
    else:
        assert jobs.job_documentation("ECL2CSV") is None
        assert jobs.job_documentation("CSV2ECL") is None

    assert jobs.job_documentation("foobar") is None


def test_get_module_variable():
    """Test that we can robustly peek into jobs for metadata.

    This is independent whether ERT is installed or not
    """
    assert jobs._get_module_variable_if_exists("foo", "bar") == ""
    assert jobs._get_module_variable_if_exists(
        "ecl2df.ecl2csv", "DESCRIPTION"
    ).startswith("Convert Eclipse input and output")
    assert jobs._get_module_variable_if_exists("ecl2df.ecl2csv", "NOPE") == ""


@pytest.mark.skipif(HAVE_ERT, reason="Tested only when ERT is not available")
def test_no_erthooks():
    """Test that we can import the hook implementations even when ERT is unavailable."""
    from ecl2df.hook_implementations import jobs  # noqa



"""Constants for use in ecl2df."""

# This is a magic filename that means read/write from/to stdout
# This makes it impossible to write to a file called "-" on disk
# but that would anyway create a lot of other problems in the shell.
MAGIC_STDOUT: str = "-"



import os
from pathlib import Path

import pytest

from ecl2df import EclFiles

try:
    import opm  # noqa

    HAVE_OPM = True
except ImportError:
    HAVE_OPM = False


TESTDIR = Path(__file__).absolute().parent
EIGHTCELLS = str(TESTDIR / "data/eightcells/EIGHTCELLS.DATA")


@pytest.mark.skipif(not HAVE_OPM, reason="OPM is not installed")
def test_filedescriptors():
    """Test that filedescriptors are properly closed"""

    fd_dir = Path("/proc/") / str(os.getpid()) / "fd"
    if not fd_dir.exists():
        print("Counting file descriptors on non-Linux not supported")
        return

    pre_fd_count = len(list(fd_dir.glob("*")))

    eclfiles = EclFiles(EIGHTCELLS)
    # No opened files yet:
    assert len(list(fd_dir.glob("*"))) == pre_fd_count

    eclfiles.close()
    # No change, no files to close:
    assert len(list(fd_dir.glob("*"))) == pre_fd_count

    eclfiles.get_egrid()
    # This should not leave any file descriptor open
    assert len(list(fd_dir.glob("*"))) == pre_fd_count

    eclfiles.get_initfile()
    assert len(list(fd_dir.glob("*"))) == pre_fd_count
    assert eclfiles._initfile is not None
    eclfiles.close()
    assert len(list(fd_dir.glob("*"))) == pre_fd_count
    assert eclfiles._initfile is None

    eclfiles.get_rstfile()
    # Automatically closed by libecl
    assert len(list(fd_dir.glob("*"))) == pre_fd_count
    assert eclfiles._rstfile is not None
    eclfiles.close()
    assert len(list(fd_dir.glob("*"))) == pre_fd_count
    assert eclfiles._rstfile is None

    eclfiles.get_eclsum()
    assert len(list(fd_dir.glob("*"))) == pre_fd_count + 1
    eclfiles.close()
    assert len(list(fd_dir.glob("*"))) == pre_fd_count

    eclfiles.get_egridfile()
    assert len(list(fd_dir.glob("*"))) == pre_fd_count
    assert eclfiles._egridfile is not None
    eclfiles.close()
    assert len(list(fd_dir.glob("*"))) == pre_fd_count
    assert eclfiles._egridfile is None

    eclfiles.get_rftfile()
    assert len(list(fd_dir.glob("*"))) == pre_fd_count
    assert eclfiles._rftfile is not None
    eclfiles.close()
    assert len(list(fd_dir.glob("*"))) == pre_fd_count
    assert eclfiles._rftfile is None

    eclfiles.get_ecldeck()
    # This should not leave any file descriptor open
    assert len(list(fd_dir.glob("*"))) == pre_fd_count



"""Test module for user API for ecl2df"""

from pathlib import Path

import pytest

import ecl2df

try:
    import opm  # noqa

    HAVE_OPM = True
except ImportError:
    HAVE_OPM = False

TESTDIR = Path(__file__).absolute().parent
REEK = str(TESTDIR / "data/reek/eclipse/model/2_R001_REEK-0.DATA")


@pytest.mark.skipif(not HAVE_OPM, reason="Test requires OPM")
def test_userapi():
    """Test that we can act as human API user

    Functionality should be extensively tested in other code, but this is here
    to illustrate how a user could work, and ensure that it works.

    To the user reading the source: Skip all 'assert' lines, read the rest.

    """
    eclfiles = ecl2df.EclFiles(REEK)

    compdatdf = ecl2df.compdat.df(eclfiles)
    equil = ecl2df.equil.df(eclfiles)
    faults = ecl2df.faults.df(eclfiles)
    fipreports = ecl2df.fipreports.df(eclfiles)
    grid_df = ecl2df.grid.df(eclfiles)
    grst_df = ecl2df.grid.df(eclfiles, rstdates="last")
    gruptree = ecl2df.gruptree.df(eclfiles)
    nnc = ecl2df.nnc.df(eclfiles)
    pillars = ecl2df.pillars.df(eclfiles)
    rft = ecl2df.rft.df(eclfiles)
    satfunc = ecl2df.satfunc.df(eclfiles)
    smry = ecl2df.summary.df(eclfiles, datetime=True)
    trans = ecl2df.trans.df(eclfiles)
    wcon = ecl2df.wcon.df(eclfiles)

    assert "PORV" in grid_df
    assert "SOIL" not in grid_df
    assert "SOIL" in grst_df
    assert "PORV" in grst_df

    # Make some HCPV calculations
    grst_df["OILPV"] = grst_df["SOIL"] * grst_df["PORV"]
    grst_df["HCPV"] = (1 - grst_df["SWAT"]) * grst_df["PORV"]

    hcpv_table = grst_df.groupby("FIPNUM").sum()[["OILPV", "HCPV"]]
    assert not hcpv_table.empty

    # Print the HCPV table by FIPNUM:
    print()
    print((hcpv_table / 1e6).round(2))

    assert not compdatdf.empty
    assert not equil.empty
    assert not faults.empty
    assert not fipreports.empty
    assert not gruptree.empty
    assert not nnc.empty
    assert not pillars.empty
    assert not rft.empty
    assert not satfunc.empty
    assert not smry.empty
    assert not trans.empty
    assert not wcon.empty



#!/usr/bin/env python
"""
Convert dataframes (in ecl2df format) to Eclipse include files,
for selected keywords
"""

import argparse
import sys

from ecl2df import __version__, equil, pvt, satfunc, summary

# String constants in use for generating ERT forward model documentation:
DESCRIPTION: str = """Convert CSV files into Eclipse include files. Uses the command
line utility ``csv2ecl``. Run ``csv2ecl --help`` to see which subcommands are supported.
No options other than the output file is possible when used directly as a forward model.
When writing synthetic summary files, the ECLBASE with no filename suffix is expected
as the OUTPUT argument."""
CATEGORY: str = "utility.eclipse"
EXAMPLES: str = (
    "``FORWARD_MODEL "
    "CSV2ECL(<SUBCOMMAND>=equil, <CSVFILE>=equil.csv, "
    "<OUTPUT>=eclipse/include/equil.inc)``"
    "CSV2ECL(<SUBCOMMAND>=summary, <CSVFILE>=summary-monthly.csv, "
    "<OUTPUT>=eclipse/model/MONTHLYSUMMARY)``"
)


def get_parser() -> argparse.ArgumentParser:
    """Make parser"""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        description=(
            "csv2ecl (" + __version__ + ") is a command line frontend to ecl2df. "
            "Documentation at https://equinor.github.io/ecl2df/ "
        ),
    )
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s {__version__}",
    )

    if sys.version_info.major >= 3 and sys.version_info.minor >= 7:
        subparsers = parser.add_subparsers(  # type: ignore
            required=True,
            dest="subcommand",
            parser_class=argparse.ArgumentParser,
        )
    else:
        subparsers = parser.add_subparsers(parser_class=argparse.ArgumentParser)

    summary_parser = subparsers.add_parser(
        "summary",
        help="Write EclSum UNSMRY files",
        description=("Write Eclipse UNSMRY files from CSV files."),
    )
    summary.fill_reverse_parser(summary_parser)
    summary_parser.set_defaults(func=summary.summary_reverse_main)

    equil_parser = subparsers.add_parser(
        "equil",
        help="Write SOLUTION include files",
        description=(
            "Write SOLUTION keywords (EQUIL, RSVD, RVVD) "
            "to Eclipse include files from CSV in ecl2df format."
        ),
    )
    equil.fill_reverse_parser(equil_parser)
    equil_parser.set_defaults(func=equil.equil_reverse_main)

    pvt_parser = subparsers.add_parser(
        "pvt",
        help="Write PVT include files",
        description=(
            "Write Eclipse include files from CSV files on the ecl2df format."
        ),
    )
    pvt.fill_reverse_parser(pvt_parser)
    pvt_parser.set_defaults(func=pvt.pvt_reverse_main)

    satfunc_parser = subparsers.add_parser(
        "satfunc",
        help="Write saturation function include files",
        description=(
            "Write saturation function include files from CSV files on "
            "the ecl2df format."
        ),
    )
    satfunc.fill_reverse_parser(satfunc_parser)
    satfunc_parser.set_defaults(func=satfunc.satfunc_reverse_main)

    return parser


def main() -> None:
    """Entry point"""
    parser = get_parser()
    args = parser.parse_args()
    args.func(args)


if __name__ == "__main__":
    main()



"""Test installation"""

import subprocess

import pytest

import ecl2df

try:
    import opm  # noqa
except ImportError:
    pytest.skip(
        "OPM is not installed, command line client does not work without it.",
        allow_module_level=True,
    )


@pytest.mark.integration
def test_integration():
    """Test that all endpoints that are to be installed are installed"""
    assert subprocess.check_output(["ecl2csv", "-h"])  # nosec
    assert subprocess.check_output(["csv2ecl", "-h"])  # nosec

    # The subparsers should exit "cleanly" with exit code 2 ("Incorrect usage")
    # when no more options are provided on the command line
    with pytest.raises(subprocess.CalledProcessError) as exception:
        subprocess.check_output(["ecl2csv"])  # nosec
        assert exception.value.returncode == 2
    with pytest.raises(subprocess.CalledProcessError) as exception:
        subprocess.check_output(["csv2ecl"])  # nosec
        assert exception.value.returncode == 2
    # ref: https://stackoverflow.com/questions/23714542/  \
    #              why-does-pythons-argparse-use-an-error-code-of-2-for-systemexit

    for submodule in ecl2df.SUBMODULES:
        helptext = subprocess.check_output(["ecl2csv", submodule, "-h"])
        # Test that this option is hidden, the argument is only there
        # to support optional number of arguments in ERT forward models.
        assert "hiddenemptyplaceholders" not in str(helptext)



#!/usr/bin/env python
"""
Extract transmissibility information from Eclipse output files as Dataframes.
"""
import argparse
import logging
from typing import List, Union

import pandas as pd

import ecl2df.grid
import ecl2df.nnc
from ecl2df import getLogger_ecl2csv
from ecl2df.common import write_dframe_stdout_file

from .eclfiles import EclFiles

try:
    import networkx

    HAVE_NETWORKX = True
except ImportError:
    HAVE_NETWORKX = False

logger = logging.getLogger(__name__)


def df(
    eclfiles: EclFiles,
    vectors: Union[str, List[str]] = None,
    boundaryfilter: bool = False,
    group: bool = False,
    coords: bool = False,
    onlykdir: bool = False,
    onlyijdir: bool = False,
    addnnc: bool = False,
) -> pd.DataFrame:
    """Make a dataframe of the neighbour transmissibilities.

    The TRANX, TRANY and TRANZ (whenever nonzero) will be used
    to produce a row representing a cell-pair where there is
    transmissibility.

    You will get a dataframe with the columns::

        I1, J1, K1, I2, J2, K2, DIR, TRAN

    similar to what you get from non-neighbour connection export.

    If you ask for coordinates, you will also get the distance (DX,
    DY, DZ) for each connection.

    The DIR column indicates the direction, and can take the
    string values I, J or K.

    If you ask for additional vectors, like FIPNUM, then
    you will get a corresponding FIPNUM1 and FIPNUM2 added.

    Args:
        eclfiles: An object representing your Eclipse run
        vectors: Eclipse INIT vectors that you want to include
        boundaryfilter: Set to true if you want to filter where one INIT
            vector change. Only use for integer INIT vectors.
        group: Set to true if you want to sum transmissibilities over
            boundary interfaces. Implies boundaryfilter and requires only one integer
            INIT vector.
        coords: Set to true if you want to add coordinates for the
            average of the two cells centerpoint and the distance between
            the two cells centerpoints (X, Y, Z, DX, DY, DZ)
        onlykdir: Set to true if you only want transmissibilities in
            K direction
        onlyijdir: Set to true if you only want transmissibilities
            in the IJ-plane
        addnnc: Set to true if NNC connection should be concatenated to
            the dataframe

    Returns:
        Dataframe with one cell-pair pr. row. Empty dataframe if error.
    """
    if not vectors:
        vectors = []
    if not isinstance(vectors, list):
        vectors = [vectors]

    if group:
        # grouping implies boundaryfilter
        boundaryfilter = True

    if boundaryfilter and len(vectors) > 1:
        logger.error(
            "Can't filter to boundaries when more than one INIT vector is supplied"
        )
        return pd.DataFrame()

    if group and len(vectors) > 1:
        logger.error("Can't group to more than one INIT vector at a time")
        return pd.DataFrame()

    if onlykdir and onlyijdir:
        logger.warning(
            "Filtering to both k and to ij simultaneously results in empty dataframe"
        )

    grid_df = ecl2df.grid.df(eclfiles)
    existing_vectors = [vec for vec in vectors if vec in grid_df.columns]
    if len(existing_vectors) < len(vectors):
        logger.warning(
            "Vectors %s not found, skipping", str(set(vectors) - set(existing_vectors))
        )
    vectors = existing_vectors
    logger.info("Building transmissibility dataframe")
    if not onlykdir:
        tranx = pd.DataFrame(grid_df[grid_df["TRANX"] > 0][["I", "J", "K", "TRANX"]])
        tranx.rename(
            columns={"I": "I1", "J": "J1", "K": "K1", "TRANX": "TRAN"}, inplace=True
        )
        tranx["I2"] = tranx["I1"] + 1
        tranx["J2"] = tranx["J1"]
        tranx["K2"] = tranx["K1"]
        tranx["DIR"] = "I"
    else:
        tranx = pd.DataFrame()

    if not onlykdir:
        trany = pd.DataFrame(grid_df[grid_df["TRANY"] > 0][["I", "J", "K", "TRANY"]])
        trany.rename(
            columns={"I": "I1", "J": "J1", "K": "K1", "TRANY": "TRAN"}, inplace=True
        )
        trany["I2"] = trany["I1"]
        trany["J2"] = trany["J1"] + 1
        trany["K2"] = trany["K1"]
        trany["DIR"] = "J"
    else:
        trany = pd.DataFrame()

    if not onlyijdir:
        tranz = pd.DataFrame(grid_df[grid_df["TRANZ"] > 0][["I", "J", "K", "TRANZ"]])
        tranz.rename(
            columns={"I": "I1", "J": "J1", "K": "K1", "TRANZ": "TRAN"}, inplace=True
        )
        tranz["I2"] = tranz["I1"]
        tranz["J2"] = tranz["J1"]
        tranz["K2"] = tranz["K1"] + 1
        tranz["DIR"] = "K"
    else:
        tranz = pd.DataFrame()

    trans_df = pd.concat([tranx, trany, tranz], axis=0, sort=False)

    if addnnc:
        logger.info("Adding NNC data")
        nnc_df = ecl2df.nnc.df(eclfiles, coords=False, pillars=False)
        nnc_df["DIR"] = "NNC"
        trans_df = pd.concat([trans_df, nnc_df], sort=False)

    # If we have additional vectors we want, merge them in:
    vectorscoords = list(vectors)  # Copy
    if coords:
        if "X" not in vectorscoords:
            vectorscoords.append("X")
        if "Y" not in vectorscoords:
            vectorscoords.append("Y")
        if "Z" not in vectorscoords:
            vectorscoords.append("Z")

    if vectorscoords:
        logger.info("Adding vectors %s", str(vectorscoords))
        grid_df = grid_df.reset_index()
        trans_df = pd.merge(
            trans_df,
            grid_df[["I", "J", "K"] + vectorscoords],
            left_on=["I1", "J1", "K1"],
            right_on=["I", "J", "K"],
        )
        trans_df = trans_df.drop(["I", "J", "K"], axis=1)
        trans_df = pd.merge(
            trans_df,
            grid_df[["I", "J", "K"] + vectorscoords],
            left_on=["I2", "J2", "K2"],
            right_on=["I", "J", "K"],
            suffixes=("1", "2"),
        )
        trans_df = trans_df.drop(["I", "J", "K"], axis=1)

    if coords:
        trans_df["X"] = (trans_df["X1"] + trans_df["X2"]) / 2.0
        trans_df["Y"] = (trans_df["Y1"] + trans_df["Y2"]) / 2.0
        trans_df["Z"] = (trans_df["Z1"] + trans_df["Z2"]) / 2.0
        trans_df["DX"] = abs(trans_df["X1"] - trans_df["X2"])
        trans_df["DY"] = abs(trans_df["Y1"] - trans_df["Y2"])
        trans_df["DZ"] = abs(trans_df["Z1"] - trans_df["Z2"])
        trans_df = trans_df.drop(["X1", "X2", "Y1", "Y2", "Z1", "Z2"], axis=1)

    if boundaryfilter:
        assert len(vectors) == 1
        logger.info(
            "Filtering to transmissibilities crossing different %s values", vectors[0]
        )
        vec1 = vectors[0] + "1"
        vec2 = vectors[0] + "2"
        trans_df = trans_df[(trans_df[vec1] != trans_df[vec2])]

    if group:
        assert len(vectors) == 1  # This is checked above
        assert boundaryfilter
        logger.info("Grouping transmissiblity over %s interfaces", str(vectors[0]))
        vec1 = vectors[0] + "1"
        vec2 = vectors[0] + "2"
        pairname = vectors[0] + "PAIR"
        # Construct a column with values like "3-4" for a pair
        # where FIPNUM1 is 4 and FIPNUM2 is 3
        trans_df[pairname] = [
            str(int(min(x[1:3]))) + "-" + str(int(max(x[1:3])))
            for x in trans_df[[vec1, vec2]].itertuples()
        ]

        aggregators = {
            "X": "mean",
            "Y": "mean",
            "Z": "mean",
            "DX": "mean",
            "DY": "mean",
            "DZ": "mean",
            "TRAN": "sum",
        }
        aggregators = {
            key: value for (key, value) in aggregators.items() if key in trans_df
        }
        trans_df = trans_df.groupby(pairname).agg(aggregators).reset_index()

        # Reinstate FIPNUM1 and FIPNUM2 (by splitting the pair, to get sorting for free)
        vectuples = trans_df[pairname].str.split("-")
        trans_df[vec1] = [tup[0] for tup in vectuples]
        trans_df[vec2] = [tup[1] for tup in vectuples]

    return trans_df


def make_nx_graph(eclfiles: EclFiles, region: str = "FIPNUM") -> "networkx.Graph":
    """Construct a networkx graph for the transmissibilities."""
    if not HAVE_NETWORKX:
        logger.error("Please install networkx for this function to work")
        return None
    trans_df = df(eclfiles, vectors=[region], coords=True, group=True)
    reg1 = region + "1"
    reg2 = region + "2"
    graph = networkx.Graph()
    graph.add_weighted_edges_from(
        [tuple(row) for row in trans_df[[reg1, reg2, "TRAN"]].values]
    )
    return graph


def fill_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Set up sys.argv parser.

    Arguments:
        parser: argparse.ArgumentParser or argparse.subparser
    """
    parser.add_argument(
        "DATAFILE",
        help="Name of Eclipse DATA file. " + "INIT and EGRID file must lie alongside.",
    )
    parser.add_argument("--vectors", nargs="+", help="Extra INIT vectors to be added")
    parser.add_argument(
        "--boundaryfilter",
        action="store_true",
        help=(
            "Filter to connections where the INIT vector change value. "
            "Only one INIT vector allowed."
        ),
    )
    parser.add_argument(
        "--onlyk", action="store_true", help="Filter to only K direction"
    )
    parser.add_argument("--onlyij", action="store_true", help="Filter to only IJ-plane")
    parser.add_argument(
        "--coords", action="store_true", help="Add coordinates to dataframe"
    )
    parser.add_argument(
        "--group",
        action="store_true",
        help=(
            "Group transmissibilities over region "
            "interfaces. Specify the region name in --vectors"
        ),
    )
    parser.add_argument(
        "--nnc",
        action="store_true",
        help="Add NNC transmissibilities to the same dataframe",
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="Be verbose")
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="Name of output csv file. Use '-' for stdout",
        default="trans.csv",
    )
    return parser


def trans_main(args):
    """This is the command line API"""
    logger = getLogger_ecl2csv(  # pylint: disable=redefined-outer-name
        __name__, vars(args)
    )
    eclfiles = EclFiles(args.DATAFILE)
    trans_df = df(
        eclfiles,
        vectors=args.vectors,
        boundaryfilter=args.boundaryfilter,
        onlykdir=args.onlyk,
        onlyijdir=args.onlyij,
        coords=args.coords,
        group=args.group,
        addnnc=args.nnc,
    )

    write_dframe_stdout_file(trans_df, args.output, index=False, caller_logger=logger)



"""Test module for parameters"""

import json
import os
from pathlib import Path

import pytest
import yaml

from ecl2df.eclfiles import EclFiles
from ecl2df.parameters import find_parameter_files, load, load_all

TESTDIR = Path(__file__).absolute().parent
DATAFILE = str(TESTDIR / "data/reek/eclipse/model/2_R001_REEK-0.DATA")


def test_parameters():
    """Test import of parameters.txt++"""
    eclfiles = EclFiles(DATAFILE)

    # NB: This test easily fails due to remnants of other test code..
    assert not find_parameter_files(eclfiles)

    parameterstxt = Path(eclfiles.get_path()) / "parameters.txt"
    # If this exists, it is a remnant from test code that has
    # crashed. It should NOT be in git.
    if parameterstxt.is_file():
        parameterstxt.unlink()
    parameterstxt.write_text("FOO 1\nBAR 3", encoding="utf-8")
    assert Path(parameterstxt).is_file()
    param_dict = load(parameterstxt)
    assert "FOO" in param_dict
    assert "BAR" in param_dict

    assert len(find_parameter_files(eclfiles)) == 1
    parameterstxt.unlink()

    parameterstxt = Path(eclfiles.get_path()).parent / "parameters.txt"
    if parameterstxt.is_file():
        parameterstxt.unlink()
    parameterstxt.write_text("FOO 1\nBAR 3\nCONTACT:BARF 2700", encoding="utf-8")
    assert Path(parameterstxt).is_file()
    param_dict = load(parameterstxt)
    assert "FOO" in param_dict
    assert "BAR" in param_dict
    assert param_dict["BAR"] == 3
    assert param_dict["CONTACT:BARF"] == 2700
    assert len(find_parameter_files(eclfiles)) == 1
    parameterstxt.unlink()

    # Typical parameters.json structure: The group "CONTACT" is assumed having
    # duplicate information, and is to be ignored
    dump_me = {"FOO": 1, "BAR": "com", "CONTACT:BARF": 2700, "CONTACT": {"BARF": 2700}}

    parametersyml = Path(eclfiles.get_path()) / "parameters.yml"
    if parametersyml.is_file():
        parametersyml.unlink()
    parametersyml.write_text(yaml.dump(dump_me), encoding="utf-8")
    assert Path(parametersyml).is_file()
    assert len(find_parameter_files(eclfiles)) == 1
    param_dict = load(parametersyml)
    assert "FOO" in param_dict
    assert "BAR" in param_dict
    assert param_dict["BAR"] == "com"
    parametersyml.unlink()

    parametersjson = Path(eclfiles.get_path()) / "parameters.json"
    if parametersjson.is_file():
        parametersjson.unlink()
    parametersjson.write_text(json.dumps(dump_me), encoding="utf-8")
    assert Path(parametersjson).is_file()
    assert len(find_parameter_files(eclfiles)) == 1
    param_dict = load(find_parameter_files(eclfiles)[0])
    param_dict_m = load_all(find_parameter_files(eclfiles))
    assert "FOO" in param_dict
    assert "BAR" in param_dict
    assert param_dict["BAR"] == "com"
    assert param_dict == param_dict_m
    parametersjson.unlink()


def test_multiple_parameters():
    """Test what happens when we have duplicate parameter files"""
    eclfiles = EclFiles(DATAFILE)
    parametersjson = Path(eclfiles.get_path()) / "parameters.json"
    parameterstxt = Path(eclfiles.get_path()).parent / "parameters.txt"
    parameterstxt.write_text("FOO 1\nBAR 4", encoding="utf-8")
    parametersjson.write_text(json.dumps({"BAR": 5, "COM": 6}), encoding="utf-8")
    param_dict = load_all(find_parameter_files(eclfiles))
    assert len(param_dict) == 3
    assert param_dict["BAR"] == 5  # json has precedence over txt
    parametersjson.unlink()
    parameterstxt.unlink()


def test_find_parameter_files_modeldir(tmp_path):
    """Test find_parameter_files when parameters are in the model directory."""
    os.chdir(tmp_path)
    model_dir = Path("realization-0/iter-0/eclipse/model")
    model_dir.mkdir(parents=True)
    assert find_parameter_files(model_dir / "FOO.DATA") == []
    (model_dir / "parameters.txt").touch()
    assert find_parameter_files(model_dir / "FOO.DATA") == [
        model_dir.absolute() / "parameters.txt"
    ]

    # yml are preferred over txt:
    (model_dir / "parameters.yml").touch()
    assert find_parameter_files(model_dir / "FOO.DATA") == [
        model_dir.absolute() / "parameters.yml",
        model_dir.absolute() / "parameters.txt",
    ]

    # json is preferred over yml again:
    (model_dir / "parameters.json").touch()
    assert find_parameter_files(model_dir / "FOO.DATA") == [
        model_dir.absolute() / "parameters.json",
        model_dir.absolute() / "parameters.yml",
        model_dir.absolute() / "parameters.txt",
    ]

    # Changing filebase:
    assert find_parameter_files(model_dir / "FOO.DATA", filebase="foo") == []
    (model_dir / "foo.yml").touch()
    assert find_parameter_files(model_dir / "FOO.DATA", filebase="foo") == [
        model_dir.absolute() / "foo.yml"
    ]

    with pytest.raises(TypeError):
        find_parameter_files({"foo": "bar"})


def test_find_parameter_files_verticalplacement(tmp_path):
    """Test find_parameter_files with parameters.txt placed above in the hiearchy."""
    os.chdir(tmp_path)
    model_dir = Path("foo/bar/realization-0/iter-0/eclipse/model")
    model_dir.mkdir(parents=True)

    # Too shallow parameters.txt is not found:
    (model_dir.parent.parent.parent / "parameters.txt").touch()
    assert find_parameter_files(model_dir / "FOO.DATA") == []

    # Two levels (realization-level) is found:
    (model_dir.parent.parent / "parameters.txt").touch()
    assert find_parameter_files(model_dir / "FOO.DATA") == [
        model_dir.absolute().parent.parent / "parameters.txt"
    ]

    # One level up is found:
    (model_dir.parent.parent / "parameters.txt").unlink()
    (model_dir.parent / "parameters.txt").touch()
    assert find_parameter_files(model_dir / "FOO.DATA") == [
        model_dir.absolute().parent / "parameters.txt"
    ]

    # If found both places, deepest location is preferred:
    (model_dir.parent.parent / "parameters.txt").touch()
    (model_dir.parent / "parameters.txt").touch()
    assert find_parameter_files(model_dir / "FOO.DATA") == [
        model_dir.absolute().parent / "parameters.txt",
        model_dir.absolute().parent.parent / "parameters.txt",
    ]


def test_load(tmp_path):
    """Test loading of yml/json/txt files into dictionaries"""
    os.chdir(tmp_path)

    Path("empty").touch()
    assert load("empty") == {}  # A warning is logged

    # yaml file:
    Path("foo.yml").write_text("foo: bar")
    assert load("foo.yml") == {"foo": "bar"}
    assert load(Path("foo.yml")) == {"foo": "bar"}

    # yaml syntax errors:
    Path("error.yml").write_text("foo: bar :")
    with pytest.raises(ValueError, match="Could not parse error.yml"):
        load("error.yml")

    # json file:
    Path("foo.json").write_text('{\n  "foo": "bar"\n}')
    assert load("foo.json") == {"foo": "bar"}

    # Extension does not matter:
    Path("wrongextension.yml").write_text('{\n  "foo": "bar"\n}')
    assert load("wrongextension.yml") == {"foo": "bar"}

    # txt file:
    Path("foo.txt").write_text("foo bar")
    assert load("foo.txt") == {"foo": "bar"}

    # txt file errors:
    Path("error.txt").write_text("foo bar com")
    with pytest.raises(ValueError, match="Could not parse error.txt"):
        load("error.txt") == {"foo": "bar"}

    with pytest.raises(FileNotFoundError, match="notexisting not found"):
        load("notexisting")



"""Extract statistics pr cornerpoint pillar (i,j)-pair"""

import argparse
import datetime
import logging
from typing import Dict, List, Optional, Union

import dateutil.parser
import pandas as pd

from ecl2df import EclFiles, common, getLogger_ecl2csv, grid

logger: logging.Logger = logging.getLogger(__name__)

AGGREGATORS: Dict[str, str] = {
    "VOLUME": "sum",
    "PORV": "sum",
    "PERMX": "mean",
    "PERMY": "mean",
    "PERMZ": "mean",
    "X": "mean",
    "Y": "mean",
    "Z": "mean",
    "WATVOL": "sum",
    "GASVOL": "sum",
    "OILVOL": "sum",
    "GASVOLSURF": "sum",
    "OILVOLSURF": "sum",
    "OWC": "mean",
    "GOC": "mean",
    "GWC": "mean",
}


def df(
    eclfiles: EclFiles,
    region: str = None,
    rstdates: Optional[Union[str, datetime.date, List[datetime.date]]] = None,
    soilcutoff: float = 0.2,
    sgascutoff: float = 0.7,
    swatcutoff: float = 0.7,
    stackdates: bool = False,
) -> pd.DataFrame:
    """Produce a dataframe with pillar information

    This is the "main" function for Python API users
    Produces a dataframe with data for each I-J combination
    (in the column PILLAR), and if a region parameter is
    supplied, also pr. region.

    PORV is the summed porevolume of the pillar (in the region),
    VOLUME is bulk volume, and PORO is porevolume weighted porosity
    PERM columns contain unweighted value averages, use with caution.

    If a restart date is picked, then SWAT and SGAS will
    be used to compute volumes pr. phase, WATVOL, OILVOL and GASVOL. The
    columns with dynamic data will include the date in the column headers
    like SWAT@2009-01-01

    Args:
        region: A parameter the pillars will be split
            on. Typically EQLNUM or FIPNUM. Set to empty string
            or None to avoid any region grouping.
        rstdates: Dates for which restart data
            is to be extracted. The string can
            be in ISO-format, or one of the mnenomics
            'first', 'last' or 'all'. It can also be a list
            of datetime.date.
        soilcutoff: If not None, an oil-water contact will
            be estimated pr. pillar, based on the deepest cell with
            SOIL above the given cutoff. Value is put in column OWC.
        sgascuttof: If not None, a gas contact will be
            estimated pr pillar, based on the deepest cell with
            SGAS above the given cutoff. Value is put in column GOC.
        swatcutoff: OWC or GWC is only computed for pillars
            where at least one cell is above this value.
        stackdates: If true, a column
            called DATE will be added and data for all restart
            dates will be added in a stacked manner.
    """
    # List of vectors we want, conservative in order to save memory and cputime:
    vectors = []
    if region:
        vectors.append(region)
    vectors.extend(["POR*", "PERM*", "SWAT", "SGAS", "1OVERBO", "1OVERBG"])
    grid_df = grid.df(eclfiles, rstdates=rstdates, vectors=vectors, dateinheaders=True)

    rstdates_iso = grid.dates2rstindices(eclfiles, rstdates)[2]

    grid_df["PILLAR"] = grid_df["I"].astype(str) + "-" + grid_df["J"].astype(str)
    logger.info("Computing pillar statistics")
    groupbies = ["PILLAR"]
    if region:
        if region not in grid_df:
            logger.warning("Region parameter %s not found, ignored", region)
        else:
            groupbies.append(region)
            grid_df[region] = grid_df[region].astype(int)

    for datestr in rstdates_iso:
        logger.info("Dynamic volumes for %s", datestr)
        volumes = compute_volumes(grid_df, datestr=datestr)
        grid_df = pd.concat([grid_df, volumes], axis="columns", sort=False)

    aggregators = {
        key: AGGREGATORS[key.split("@")[0]]
        for key in grid_df
        if key.split("@")[0] in AGGREGATORS
    }

    # Group over PILLAR and possibly regions:
    grouped = (grid_df.groupby(groupbies).agg(aggregators)).reset_index()

    # Compute correct pillar averaged porosity (from bulk)
    if "PORV" in grouped and "VOLUME" in grouped:
        grouped["PORO"] = grouped["PORV"] / grouped["VOLUME"]

    # Compute contacts:
    for datestr in rstdates_iso:
        if "SWAT@" + datestr in grid_df and (
            "SOIL@" + datestr in grid_df or "SGAS@" + datestr in grid_df
        ):
            contacts = compute_pillar_contacts(
                grid_df,
                region=region,
                soilcutoff=soilcutoff,
                sgascutoff=sgascutoff,
                swatcutoff=swatcutoff,
                datestr=datestr,
            )
            if not contacts.empty:
                grouped = pd.merge(grouped, contacts, how="left")

    if stackdates:
        return common.stack_on_colnames(
            grouped, sep="@", stackcolname="DATE", inplace=True
        )
    return grouped


def compute_volumes(
    grid_df: pd.DataFrame, datestr: Optional[str] = None
) -> pd.DataFrame:
    """Compute "dynamic" volumes, volumes for data coming from the
    UNRST file (SWAT+SGAS)

    SOIL is assumed not be present in grid_df, it will be computed here.

    Args:
        grid_df: A dataframe with the columns PORV, SWAT and SGAS
        datestr: If not none, it should contain an ISO-8601 formatted date that
            will be appended to all columns with dynamic data, SWAT and SGAS
            in incoming grid_df must also be appended with the same date string.
            If datestr is "2009-01-02", then the columns will look
            like "SOIL@2009-01-02"
    """
    if datestr:
        assert isinstance(dateutil.parser.parse(datestr).date(), datetime.date)
        atdatestr = "@" + str(datestr)
    else:
        atdatestr = ""
    vols = pd.DataFrame()
    if "SWAT" + atdatestr in grid_df and "SGAS" + atdatestr in grid_df:
        vols["SOIL" + atdatestr] = (
            1 - grid_df["SWAT" + atdatestr] - grid_df["SGAS" + atdatestr]
        )
    if "SWAT" + atdatestr in grid_df and "SGAS" + atdatestr not in grid_df:
        # Two-phase oil-water
        vols["SOIL" + atdatestr] = 1 - grid_df["SWAT" + atdatestr]
        # (or it could be two-phase water-gas, but then the SOIL would mean the same)

    if "SWAT" + atdatestr in grid_df:
        vols["WATVOL" + atdatestr] = grid_df["SWAT" + atdatestr] * grid_df["PORV"]
    if "SGAS" + atdatestr in grid_df:
        vols["GASVOL" + atdatestr] = grid_df["SGAS" + atdatestr] * grid_df["PORV"]
    if "SOIL" + atdatestr in vols:
        vols["OILVOL" + atdatestr] = vols["SOIL" + atdatestr] * grid_df["PORV"]

    if "1OVERBO" + atdatestr in grid_df and "OILVOL" + atdatestr in vols:
        vols["OILVOLSURF" + atdatestr] = (
            vols["OILVOL" + atdatestr] * grid_df["1OVERBO" + atdatestr]
        )

    if "1OVERBG" + atdatestr in grid_df and "GASVOL" + atdatestr in vols:
        vols["GASVOLSURF" + atdatestr] = (
            vols["GASVOL" + atdatestr] * grid_df["1OVERBG" + atdatestr]
        )
    return vols


def compute_pillar_contacts(
    grid_df: pd.DataFrame,
    region: Optional[str] = None,
    soilcutoff: float = 0.2,
    sgascutoff: float = 0.7,
    swatcutoff: float = 0.7,
    datestr: Optional[str] = None,
) -> pd.DataFrame:
    """Compute contacts pr. pillar in a grid dataframe.

    Requires the columns PILLAR, SOIL, SGAS and SWAT, I, J and Z

    SOIL should be pre-computed in three-phase runs before calling this.
    PILLAR must be pre-computed in grid_df.

    OWC is deepest cell centre pr. pillar with oil saturation above soilcutoff,
    among those pillars with at least one cell above swatcutoff.

    GOC is the deepest cell centre pr. pillar with gas saturation above sgascutoff,
    among those pillars with at least one cell with non-zero oil saturation

    GWC is only computed when SOIL is not presented, and is the deepest
    cell centre with gas saturation above sgascutoff, among those pillars
    with at least one cell above swatcutoff.

    Args:
        grid_df: Representing the grid with geometry and properties
        region: Region vector to group by, e.g. FIPNUM or EQLNUM
        soilcutoff: OWC is deepest cell centre pr. pillar with oil
            saturation above this
        sgascutoff: GOC/GWC is deepest cell centre pr. pillar with
            gas saturation above this
        swatcutoff: Pillars must have this amount of water in (in one cell)
            them to be available for OWC/GWC computations.
    Returns:
        Index is PILLAR with values I-J. Rows only for
        pillars where a contact was found. Empty dataframe if no contacts found.
    """
    assert 0 <= swatcutoff <= 1
    assert 0 <= soilcutoff <= 1
    assert 0 <= sgascutoff <= 1

    # assert datestr is None or in ISO-8601 format
    if datestr:
        atdatestr = "@" + datestr
    else:
        atdatestr = ""

    # Non-user servicable parameter, for GOC computation
    # we require cells at the contact to contain a minute saturation
    # of oil. This is to avoid the code picking up gas injected in the
    # water phase.
    epsilon_soil = 0.01

    if "SWAT" + atdatestr not in grid_df:
        logger.warning("No saturation in grid data. No contacts computed")
        return pd.DataFrame()
    logger.info("Computing contacts pr. pillar")
    groupbies = ["PILLAR"]
    if "PILLAR" not in grid_df:
        grid_df["PILLAR"] = grid_df["I"].astype(str) + "-" + grid_df["J"].astype(str)

    if "Z" not in grid_df:
        # To ensure same exception across Python 3.x:
        raise KeyError("Z column must be present in dataframe")

    if region:
        groupbies.append(region)
    owc = pd.DataFrame()
    goc = pd.DataFrame()

    # Extract pillars with water in them. Only here would an OWC make sense.
    waterpillars = (
        grid_df[grid_df["SWAT" + atdatestr] > swatcutoff]
        .groupby(groupbies)
        .agg({"Z": "min"})
        .reset_index()
    )
    if soilcutoff and "SOIL" + atdatestr in grid_df:
        logger.info(
            "Calculating oil-water-contacts based on SOILcutoff %s", str(soilcutoff)
        )
        owc = (
            grid_df[grid_df["SOIL" + atdatestr] > soilcutoff]
            .groupby(groupbies)
            .agg({"Z": "max"})
        )
        owc.rename(columns={"Z": "OWC" + atdatestr}, inplace=True)
        owc.reset_index(inplace=True)
        # Filter the owc frame to only those pillars that also has water:
        owc = pd.merge(waterpillars, owc, how="inner").drop("Z", axis="columns")

    if sgascutoff and "SGAS" + atdatestr in grid_df:
        logger.info("Calculating gas-contacts based on gas cutoff %s", str(sgascutoff))
        if "SOIL" + atdatestr in grid_df and "SGAS" + atdatestr in grid_df:
            # Pillars to be used for GOC computation
            gocpillars = (
                grid_df[grid_df["SOIL" + atdatestr] > epsilon_soil]
                .groupby(groupbies)
                .agg({"Z": "first"})  # The actual aggregation is not used.
                .reset_index()
            )
            goc = (
                grid_df[
                    (grid_df["SGAS" + atdatestr] > sgascutoff)
                    & (grid_df["SOIL" + atdatestr] > epsilon_soil)
                ]
                .groupby(groupbies)
                .agg({"Z": "max"})
            )
            goc.rename(columns={"Z": "GOC" + atdatestr}, inplace=True)
        else:
            # Two-phase gas-water: GWC computation
            gocpillars = waterpillars  # In case of gas-water
            goc = (
                grid_df[grid_df["SGAS" + atdatestr] > sgascutoff]
                .groupby(groupbies)
                .agg({"Z": "max"})
            )
            goc.rename(columns={"Z": "GWC" + atdatestr}, inplace=True)
        goc.reset_index(inplace=True)
        # Filter the goc frame to only those with oil or water:
        goc = pd.merge(gocpillars, goc, how="inner").drop("Z", axis="columns")

    # We need to avoid merging with potentially empty DataFrames
    if owc.empty and goc.empty:
        return pd.DataFrame()
    if not owc.empty and goc.empty:
        return owc
    if owc.empty and not goc.empty:
        return goc
    return pd.merge(owc, goc)


def fill_parser(parser: argparse.ArgumentParser) -> argparse.ArgumentParser:
    """Set up sys.argv parser.

    Arguments:
        parser: argparse.ArgumentParser or argparse.subparser
    """
    parser.add_argument(
        "DATAFILE",
        help=("Name of Eclipse DATA file. " "INIT and EGRID file must lie alongside."),
    )
    parser.add_argument(
        "--region",
        help=(
            "Name of Eclipse region parameter for which to separate the computations. "
            "Set to empty string to have no grouping (only by pillar)."
        ),
        type=str,
        default="",
    )
    parser.add_argument(
        "--rstdates",
        type=str,
        help=(
            "Point in time to grab restart data from, "
            "either 'first' or 'last', 'all' or a date in "
            "YYYY-MM-DD format"
        ),
        default="",
    )
    parser.add_argument(
        "--stackdates",
        action="store_true",
        help=(
            "If set, the dates from restart data will not be in the column "
            "but instead there will be a DATE column with the dates. Note "
            "that the static data will be repeated for each DATE."
        ),
    )
    parser.add_argument(
        "--soilcutoff",
        type=float,
        help=(
            "If supplied as float, an oil-water contact will "
            "be estimated pr. pillar, based on the deepest cell with "
            "SOIL above the given cutoff. Value is put in column OWC."
        ),
        default=0.5,
    )
    parser.add_argument(
        "--sgascutoff",
        type=float,
        help=(
            "If supplied, a gas contact will be "
            "estimated pr pillar, based on the deepest cell with "
            "SGAS above the given cutoff. Value is put in column GOC. "
        ),
        default=0.5,
    )
    parser.add_argument(
        "--swatcutoff",
        type=float,
        help=(
            "For OWC or GWC computations, only pillars with at least one cell "
            "with water saturation above this value will be considered."
        ),
        default=0.5,
    )
    parser.add_argument(
        "--group",
        action="store_true",
        help=(
            "If set, output will not be pr. pillar, but grouped over "
            "all pillars. If --region is set, data will be grouped over that vector. "
            "The aggregation operator is sum or mean, depending on datatype."
        ),
    )
    parser.add_argument(
        "-o",
        "--output",
        type=str,
        help="Name of output csv file.",
        default="pillars.csv",
    )
    parser.add_argument("-v", "--verbose", action="store_true", help="Be verbose")
    return parser


def pillars_main(args) -> None:
    """This is the command line API"""
    logger = getLogger_ecl2csv(  # pylint: disable=redefined-outer-name
        __name__, vars(args)
    )

    eclfiles = EclFiles(args.DATAFILE)
    dframe = df(
        eclfiles,
        region=args.region,
        rstdates=args.rstdates,
        soilcutoff=args.soilcutoff,
        sgascutoff=args.sgascutoff,
        swatcutoff=args.swatcutoff,
        stackdates=args.stackdates,
    )
    groupbies = []
    aggregators = {
        key: AGGREGATORS[key.split("@")[0]]
        for key in dframe
        if key.split("@")[0] in AGGREGATORS
    }
    if args.region and args.group:
        groupbies.append(args.region)
    if args.stackdates and args.group:
        groupbies.append("DATE")
    if groupbies:
        dframe = dframe.groupby(groupbies).agg(aggregators).reset_index()
    elif args.group:
        dframe = dframe.drop("PILLAR", axis=1).mean().to_frame().transpose()
    dframe["PORO"] = dframe["PORV"] / dframe["VOLUME"]
    common.write_dframe_stdout_file(
        dframe, args.output, index=False, caller_logger=logger
    )
